import argparse
import os
import random
import sys

os.environ["KMP_DUPLICATE_LIB_OK"]="TRUE"
try:
    from typing import Literal
except:
    from typing_extensions import Literal
try:
    import torch
    from ultralytics import YOLO
except ModuleNotFoundError:
    YOLO = None
    torch = None

import random
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from simba.data_processors.cuda.utils import _is_cuda_available
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists, check_instance, check_int,
                                check_valid_boolean, check_valid_lst,
                                check_valid_tuple, get_fn_ext)
from simba.utils.enums import Options
from simba.utils.errors import (CountError, InvalidFileTypeError,
                                SimBAGPUError, SimBAPAckageVersionError)
from simba.utils.printing import SimbaTimer
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_video_meta_data)
from simba.utils.warnings import FileExistWarning, NoDataFoundWarning
from simba.utils.yolo import (_get_undetected_obs, filter_yolo_keypoint_data,
                              load_yolo_model, yolo_predict)

OUT_COLS = ['FRAME', 'CLASS_ID', 'CLASS_NAME', 'CONFIDENCE', 'X1', 'Y1', 'X2', 'Y2', 'X3', 'Y3', 'X4', 'Y4']
COORD_COLS = ['X1', 'Y1', 'X2', 'Y2', 'X3', 'Y3', 'X4', 'Y4']
NEAREST = 'nearest'


class YOLOPoseInference():

    """
    YOLOPoseInference performs pose estimation on videos using a YOLO-based keypoint detection model.

    This class runs YOLO-based keypoint detection on a given video or list of videos. It supports GPU acceleration,
    batch or stream-based inference, result interpolation, and saving results to disk. The model returns detected
    keypoints and their confidence scores for each frame, and optionally tracks poses over time.

    .. seealso::
       For bounding box inference only (no pose), see :func:`~simba.bounding_box_tools.yolo.yolo_inference.YoloInference`.
       For segmentation inference, see :func:`~simba.bounding_box_tools.yolo.yolo_seg_inference.YOLOSegmentationInference`.

    :param Union[str, os.PathLike] weights: Path to the trained YOLO model weights (e.g., 'best.pt').
    :param Union[str, os.PathLike] or List[Union[str, os.PathLike]] video_path: Path to a single video, list of videos, or directory containing video files.
    :param Tuple[str, ...] keypoint_names: Tuple containing the names of keypoints to be tracked (e.g., ('nose', 'left_ear', ...)). If None, ('BP_0', 'BP_1', ...) will be used.
    :param Optional[bool] verbose: If True, outputs progress information and timing. Defaults to True.
    :param Optional[Union[str, os.PathLike]] save_dir: Directory to save the inference results. If None, results are returned in memory. Defaults to None.
    :param Union[Literal['cpu'], int] device: Device to use for inference. Use 'cpu' for CPU or GPU index (e.g., 0 for CUDA:0). Defaults to 0.
    :param Optional[str] format: Optional export format for the model. Supported values: "onnx", "engine", "torchscript", "onnxsimplify", "coreml", "openvino", "pb", "tf", "tflite". Defaults to None.
    :param Optional[int] batch_size: Number of frames to process in parallel. Defaults to 4.
    :param int torch_threads: Number of PyTorch threads to use. Defaults to 8.
    :param bool half_precision: If True, uses half-precision (FP16) inference. Defaults to True.
    :param bool stream: If True, processes frames one-by-one in a generator style. Recommended for long videos. Defaults to False.
    :param float box_threshold: Confidence threshold bounding box detection. All detections (bounding boxes AND keypoints) below this value are ignored. Defaults to 0.5.
    :param Optional[int] max_tracks: Maximum number of pose tracks to keep. If None, all tracks are retained.
    :param bool interpolate: If True, interpolates missing keypoints across frames using the 'nearest' method. Defaults to False.
    :param bool overwrite: If True, overwrites the data at the ``save_dir``. If False, skips the file if it exists.
    :param bool raise_error: If True, raise error if the input video metadata can't be read. If False, then skips the video file.
    :param bool randomize_order: If True, analyzes the input data in a random order. If False, then in fixed order.
    :param int imgsz: Input image size for inference. Must be square. Defaults to 640.
    """

    def __init__(self,
                 weights: Union[str, os.PathLike, YOLO],
                 video_path: Union[Union[str, os.PathLike], List[Union[str, os.PathLike]]],
                 keypoint_names: Optional[Tuple[str, ...]] = None,
                 verbose: Optional[bool] = True,
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 device: Union[Literal['cpu'], int] = 0,
                 format: Optional[str] = None,
                 batch_size: Optional[int] = 4,
                 torch_threads: int = 8,
                 half_precision: bool = True,
                 stream: bool = False,
                 box_threshold: float = 0.5,
                 max_tracks: Optional[int] = None,
                 interpolate: bool = False,
                 imgsz: int = 640,
                 iou: float = 0.5,
                 overwrite: bool = True,
                 raise_error: bool = True,
                 randomize_order: bool = False):


        gpu_available, gpus = _is_cuda_available()
        if not gpu_available:
            raise SimBAGPUError(msg='No GPU detected.', source=self.__class__.__name__)
        else:
            print(f'GPUS AVAILABLE: {gpus}')
        if YOLO is None:
            raise SimBAPAckageVersionError(msg='ultralytics.YOLO package not detected.', source=self.__class__.__name__)
        check_valid_boolean(value=raise_error, source=f'{self.__class__.__name__} raise_error')
        if isinstance(video_path, list):
            check_valid_lst(data=video_path, source=f'{self.__class__.__name__} video_path', valid_dtypes=(str, np.str_,), min_len=1)
        elif os.path.isfile(video_path):
            video_path = [video_path]
        elif os.path.isdir(video_path):
            video_path = find_files_of_filetypes_in_directory(directory=video_path, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value, as_dict=False)
        for i in video_path:
            _ = get_video_meta_data(video_path=i, raise_error=raise_error)


        check_instance(source=f'{self.__class__.__name__} weights', instance=weights, accepted_types=(str, os.PathLike, YOLO))
        if not isinstance(weights, YOLO):
            check_file_exist_and_readable(file_path=weights)
            self.model = load_yolo_model(weights_path=weights, verbose=verbose, device=device, format=format)
        else:
            self.model = weights
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose')
        check_valid_boolean(value=interpolate, source=f'{self.__class__.__name__} interpolate')
        check_valid_boolean(value=overwrite, source=f'{self.__class__.__name__} overwrite')
        check_valid_boolean(value=randomize_order, source=f'{self.__class__.__name__} randomize_order')
        if randomize_order:
            random.shuffle(video_path)
        check_int(name=f'{self.__class__.__name__} batch_size', value=batch_size, min_value=1)
        check_int(name=f'{self.__class__.__name__} imgsz', value=imgsz, min_value=1)
        check_float(name=f'{self.__class__.__name__} threshold', value=box_threshold, min_value=10e-6, max_value=1.0)
        check_float(name=f'{self.__class__.__name__} iou', value=iou, min_value=10e-6, max_value=1.0)
        if keypoint_names is not None:
            check_valid_tuple(x=keypoint_names, source=f'{self.__class__.__name__} keypoint_names', min_integer=1, valid_dtypes=(str,))
            if self.model.model.kpt_shape[0] != len(keypoint_names):
                raise CountError(msg=f'The YOLO model expects {self.model.model.model.head.kpt_shape[0]} keypoints but you passed {len(keypoint_names)}: {keypoint_names}', source=self.__class__.__name__)
        else:
            keypoint_names = [f'BP_{x+1}' for x in range(self.model.model.kpt_shape[0])]
        if max_tracks is not None:
            check_int(name=f'{self.__class__.__name__} max_tracks', value=max_tracks, min_value=1)
        if save_dir is not None:
            check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} save_dir')
        self.keypoint_col_names = [f'{i}_{s}'.upper() for i in keypoint_names for s in ['x', 'y', 'p']]
        self.keypoint_cord_col_names = [f'{i}_{s}'.upper() for i in keypoint_names for s in ['x', 'y']]
        OUT_COLS.extend(self.keypoint_col_names)
        COORD_COLS.extend(self.keypoint_cord_col_names)
        torch.set_num_threads(torch_threads)
        self.half_precision, self.stream, self.video_path, self.raise_error = half_precision, stream, video_path, raise_error
        self.device, self.batch_size, self.threshold, self.max_tracks, self.iou = device, batch_size, box_threshold, max_tracks, iou
        self.verbose, self.save_dir, self.imgsz, self.interpolate, self.overwrite = verbose, save_dir, imgsz, interpolate, overwrite
        if self.model.model.task != 'pose':
            raise InvalidFileTypeError(msg=f'The model {weights} is not a pose model. It is a {self.model.model.task} model', source=self.__class__.__name__)


    def run(self):
        results = {}
        class_dict = self.model.names
        timer = SimbaTimer(start=True)
        for path in self.video_path:
            _, video_name, _ = get_fn_ext(filepath=path)
            if self.save_dir:
                save_path = os.path.join(self.save_dir, f'{video_name}.csv')
                if not self.overwrite and os.path.isfile(save_path):
                    FileExistWarning(msg=f'Skipping video {video_name} (already exist and overwrite equals False) ....')
                    continue
            video_meta = get_video_meta_data(video_path=path, raise_error=self.raise_error)
            if video_meta is None and not self.raise_error:
                NoDataFoundWarning(msg=f'Skipping video {video_name} (could not read video meta data)....')
                continue
            video_out = []
            video_predictions = yolo_predict(model=self.model,
                                             source=path,
                                             half=self.half_precision,
                                             batch_size=self.batch_size,
                                             stream=self.stream,
                                             imgsz=self.imgsz,
                                             device=self.device,
                                             threshold=0.0001, #self.threshold,
                                             max_detections=self.max_tracks,
                                             verbose=False, #self.verbose,
                                             iou=self.iou)
            for frm_cnt, video_prediction in enumerate(video_predictions):
                boxes = video_prediction.obb.data if video_prediction.obb is not None else video_prediction.boxes.data
                boxes = boxes.cpu().numpy().astype(np.float32)
                keypoints = video_prediction.keypoints.data.cpu().numpy().astype(np.float32)
                print(boxes)
                print(keypoints)
                detected_classes = np.unique(boxes[:, -1]).astype(np.int32) if boxes.size > 0 else []
                for class_id, class_name in class_dict.items():
                    if class_id not in detected_classes:
                        video_out.append(_get_undetected_obs(frm_id=frm_cnt, class_id=class_id, class_name=class_name, value_cnt=(9 + (len(self.keypoint_col_names)))))
                        continue
                    cls_boxes, cls_keypoints = filter_yolo_keypoint_data(bbox_data=boxes, keypoint_data=keypoints, class_id=class_id, confidence=None, class_idx=-1, confidence_idx=None)
                    for i in range(cls_boxes.shape[0]):
                        box = np.array([cls_boxes[i][0], cls_boxes[i][1], cls_boxes[i][2], cls_boxes[i][1], cls_boxes[i][2], cls_boxes[i][3], cls_boxes[i][0], cls_boxes[i][3]]).astype(np.int32)
                        bbox = np.array([frm_cnt, cls_boxes[i][-1], class_dict[cls_boxes[i][-1]], cls_boxes[i][-2]] + list(box))
                        bbox = np.append(bbox, cls_keypoints[i].flatten())
                        video_out.append(bbox)
            results[video_name] = pd.DataFrame(video_out, columns=OUT_COLS)
            results[video_name]['FRAME'] = results[video_name]['FRAME'].astype(np.int64)
            if self.interpolate:
                for cord_col in COORD_COLS:
                    results[video_name][cord_col] = results[video_name][cord_col].astype(np.float32).astype(np.int32).replace(to_replace=-1, value=np.nan)
                    results[video_name][cord_col] = results[video_name][cord_col].interpolate(method=NEAREST, axis=0).ffill().bfill()
            if self.save_dir:
                save_path = os.path.join(self.save_dir, f'{video_name}.csv')
                results[video_name].to_csv(save_path)
                del results[video_name]

        timer.stop_timer()
        if not self.save_dir:
            if self.verbose:
                print(f'YOLO results created', timer.elapsed_time_str)
            return results
        else:
            if self.verbose:
                print(f'YOLO results saved in {self.save_dir} directory', timer.elapsed_time_str)
            return None


# if __name__ == "__main__" and not hasattr(sys, 'ps1'):
#     parser = argparse.ArgumentParser(description="Perform YOLO-based keypoint pose estimation inference on videos.")
#     parser.add_argument('--weights_path', type=str, required=True, help='Path to the trained YOLO model weights (e.g., "best.pt").')
#     parser.add_argument('--video_path', type=str, nargs='+', required=True, help='One or more paths to video files to process. Can be a directory of a file path.')
#     parser.add_argument('--keypoint_names', type=str, nargs='+', required=True, help='List of keypoint names, e.g., nose left_ear right_ear.')
#     parser.add_argument('--verbose', action='store_true', default=False, help='Enable verbose logging.')
#     parser.add_argument('--save_dir', type=str, default=None, help='Directory to save output CSV files. If omitted, results are returned in memory.')
#     parser.add_argument('--device', type=str, default='0', help="Device to use: 'cpu' or GPU index as string (e.g., '0').")
#     parser.add_argument('--format', type=str, default=None, help='Optional export format: "onnx", "engine", "torchscript", "onnxsimplify", "coreml", "openvino", "pb", "tf", "tflite".')
#     parser.add_argument('--batch_size', type=int, default=4, help='Batch size for inference.')
#     parser.add_argument('--torch_threads', type=int, default=8, help='Number of PyTorch threads to use.')
#     parser.add_argument('--half_precision', action='store_true', help='Use half-precision (FP16) inference.')
#     parser.add_argument('--stream', action='store_true', help='Process frames in stream (one-by-one) mode.')
#     parser.add_argument('--threshold', type=float, default=0.5, help='Confidence threshold for detections (0.0 - 1.0).')
#     parser.add_argument('--max_tracks', type=int, default=None, help='Maximum number of pose tracks to retain.')
#     parser.add_argument('--interpolate', action='store_true', help='Interpolate missing keypoints across frames.')
#     parser.add_argument('--imgsz', type=int, default=640, help='Input image size (square). Default is 640.')
#     args = parser.parse_args()
#
#     keypoints_tuple = tuple(args.keypoint_names)
#     device_val = args.device if args.device == 'cpu' else int(args.device)
#     video_paths = args.video_path if len(args.video_path) > 1 else args.video_path[0]
#
#     inference = YOLOPoseInference(weights=args.weights_path,
#                                   video_path=video_paths,
#                                   keypoint_names=keypoints_tuple,
#                                   verbose=args.verbose,
#                                   save_dir=args.save_dir,
#                                   device=device_val,
#                                   format=args.format,
#                                   batch_size=args.batch_size,
#                                   torch_threads=args.torch_threads,
#                                   half_precision=args.half_precision,
#                                   stream=args.stream,
#                                   threshold=args.threshold,
#                                   max_tracks=args.max_tracks,
#                                   interpolate=args.interpolate,
#                                   imgsz=args.imgsz)
#     inference.run()


# VIDEO_DIR = '/mnt/filesystem/video_data_mp4'
# SAVE_DIR = '/mnt/filesystem/yolo_pose'
# WEIGHTS_PATH = '/mnt/filesystem/weights/best.pt'
# KEYPOINT_NAMES = ('Nose', 'Left_ear', 'Right_ear', 'Left_side', 'Center', 'Right_side', 'Tail_base', 'Tail_center', 'Tail_tip')
#
#
# i = YOLOPoseInference(weights=WEIGHTS_PATH,
#                         video_path=VIDEO_DIR,
#                         save_dir=SAVE_DIR,
#                         verbose=True,
#                         device=0,
#                         format=None,
#                         stream=True,
#                         keypoint_names=KEYPOINT_NAMES,
#                         batch_size=100,
#                         imgsz=640,
#                         interpolate=False,
#                         threshold=0.5,
#                         max_tracks=4,
#                         overwrite=False,
#                         randomize_order=True)
# i.run()

#
#
#
# # # #
# video_paths = r"D:\cvat_annotations\videos\mp4_20250624155703"
# weights_path = r"D:\cvat_annotations\frames\yolo_072125\mdl\train\weights\best.pt"
# save_dir = r"D:\cvat_annotations\frames\yolo_072125\results"
# # #
# keypoint_names = ('Nose', 'Left_ear', 'Right_ear', 'Left_side', 'Center', 'Right_side', 'Tail_base', 'Tail_center', 'Tail_tip')
# # # #
# # # #
# i = YOLOPoseInference(weights=weights_path,
#                       video_path=video_paths,
#                       save_dir=save_dir,
#                       verbose=True,
#                       device=0,
#                       format=None,
#                       stream=True,
#                       keypoint_names=keypoint_names,
#                       batch_size=100,
#                       imgsz=640,
#                       interpolate=False,
#                       box_threshold=0.5,
#                       max_tracks=4)
# i.run()
#
#
# # video_path = r"/mnt/c/troubleshooting/mitra/project_folder/videos/501_MA142_Gi_CNO_0521.mp4"
# # video_path = "/mnt/d/netholabs/yolo_videos/2025-05-28_19-46-56.mp4"
# # video_path = "/mnt/d/netholabs/yolo_videos/2025-05-28_19-50-23.mp4"
# #
# #save_dir = r'/mnt/d/netholabs/yolo_videos_out'
#
#
#
# # video_path = "/mnt/d/platea/platea_videos/videos/2C_Mouse_5-choice_MustTouchTrainingNEWFINAL_a8.mp4"
# # save_dir = r"/mnt/d/TS_DLC/yolo_kpt/results"
# # weights_path = "/mnt/d/TS_DLC/yolo_kpt/mdl/train2/weights/best.pt"
# #
# # keypoint_names = ('right_ear', 'left_body', 'nose', 'center', 'tail', 'right_body', 'left_ear')
#
#
#
# weights_path = r"D:\platea\yolo_071525\mdl\train4\weights\best.pt"
# video_path = r"D:\platea\platea_videos\videos\clipped\10B_Mouse_5-choice_MustTouchTrainingNEWFINAL_a7.mp4"
# save_dir=r"D:\platea\platea_videos\videos\yolo_results"
# #
# #
# #keypoint_names = ('Left_ear', 'Right_ear', 'Nose', 'Left_side', 'Right_side', 'Tail_base', 'Center', 'Tail_center', 'Tail_tip')
# keypoint_names = ('Left_ear', 'Right_ear', 'Nose', 'Left_side', 'Right_side', 'Center', 'Tail_base')
# #
# #
# i = YOLOPoseInference(weights_path=weights_path,
#                         video_path=video_path,
#                         save_dir=save_dir,
#                         verbose=True,
#                         device=0,
#                         format=None,
#                         stream=True,
#                         keypoint_names=keypoint_names,
#                         batch_size=500,
#                         imgsz=320,
#                         interpolate=True,
#                         threshold=0.8,
#                         retina_msk=True)
# i.run()


__author__ = "Simon Nilsson"

import os
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.ui.tkinter_functions import FileSelect
from simba.unsupervised.enums import Clustering, Unsupervised
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_keys_exist_in_dict)
from simba.utils.enums import Formats
from simba.utils.read_write import read_pickle


class PrintEmBeddingInfoPopUp(PopUpMixin, ConfigReader, UnsupervisedMixin):
    def __init__(self, config_path: Union[str, os.PathLike]):
        PopUpMixin.__init__(self, title="PRINT EMBEDDING MODEL INFO")
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        self.data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.dataset_file_selected = FileSelect(
            self.data_frm,
            "DATASET (PICKLE): ",
            lblwidth=25,
            initialdir=self.project_path,
            file_types=[("SimBA model", f"*.{Formats.PICKLE.value}")],
        )
        self.data_frm.grid(row=0, column=0, sticky=NW)
        self.dataset_file_selected.grid(row=0, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        check_file_exist_and_readable(file_path=self.dataset_file_selected.file_path)
        data = read_pickle(data_path=self.dataset_file_selected.file_path)
        if Unsupervised.DR_MODEL.value in data.keys():
            print("UMAP PARAMETERS")
            parameters = {
                **data[Unsupervised.DR_MODEL.value][Unsupervised.PARAMETERS.value]
            }
            print(parameters)
        if Clustering.CLUSTER_MODEL.value in data.keys():
            print("HDBSCAN PARAMETERS")
            parameters = {
                **data[Clustering.CLUSTER_MODEL.value][Unsupervised.PARAMETERS.value]
            }
            print(parameters)


# _ = PrintEmBeddingInfoPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import os
from typing import Any, Dict, Union

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.unsupervised.bout_aggregator import bout_aggregator
from simba.unsupervised.enums import Unsupervised
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_filepath_list_is_empty,
                                check_instance)
from simba.utils.errors import NoDataError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_pickle


class DatasetCreator(ConfigReader, UnsupervisedMixin):
    """
    Transform raw frame-wise supervised classification data into aggregated
    data for unsupervised analyses. Saves the aggergated data in to logs directory of the SimBa project.

    :param Union[str, os.PathLike] config_path: path to SimBA configparser.ConfigParser project_config.ini
    :param Dict[str, Any] settings: Attributes for which data should be included and how the data should be aggregated.

    :example:
    >>> settings = {'data_slice': 'ALL FEATURES (EXCLUDING POSE)', 'clf_slice': 'Attack', 'bout_aggregation_type': 'MEDIAN', 'min_bout_length': 66, 'feature_path': '/Users/simon/Desktop/envs/simba_dev/simba/assets/unsupervised/features.csv'}
    >>> db_creator = DatasetCreator(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/project_config.ini', settings=settings)
    >>> db_creator.run()
    """

    def __init__(self, config_path: Union[str, os.PathLike], settings: Dict[str, Any]):
        check_file_exist_and_readable(file_path=config_path)
        check_instance(
            source=f"{self.__class__.__name__} settings",
            instance=settings,
            accepted_types=(dict,),
        )
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        print(
            f"Creating unsupervised learning dataset from {len(self.machine_results_paths)} data files..."
        )
        check_if_filepath_list_is_empty(
            filepaths=self.machine_results_paths,
            error_msg="NO MACHINE LEARNING DATA FOUND in project_folder/csv/machine_results directory",
        )
        self.settings = settings
        self.clf_type, self.feature_lst = None, None
        self.save_path = os.path.join(
            self.logs_path, f"unsupervised_data_{self.datetime}.pickle"
        )
        self.log_save_path = os.path.join(
            self.logs_path, f"unsupervised_data_log_{self.datetime}.csv"
        )
        self.clf_probability_cols = [f"Probability_{x}" for x in self.clf_names]
        self.clf_cols = self.clf_names + self.clf_probability_cols

    def run(self):
        if (
            self.settings[Unsupervised.DATA_SLICE_SELECTION.value]
            == Unsupervised.ALL_FEATURES_EX_POSE.value
        ):
            self._run(drop_bps=True, user_defined=False)
        elif (
            self.settings[Unsupervised.DATA_SLICE_SELECTION.value]
            == Unsupervised.USER_DEFINED_SET.value
        ):
            self._run(drop_bps=False, user_defined=True)
        else:
            self._run(drop_bps=False, user_defined=False)

    def _read_files(self):
        raw_x_df = []
        for cnt, file_path in enumerate(self.machine_results_paths):
            _, video_name, _ = get_fn_ext(filepath=file_path)
            print(
                f"Reading {video_name} (File {cnt + 1}/{len(self.machine_results_paths)})..."
            )
            df = read_df(file_path=file_path, file_type=self.file_type)
            df.insert(0, Unsupervised.FRAME.value, df.index)
            df.insert(0, Unsupervised.VIDEO.value, video_name)
            raw_x_df.append(df)
        return pd.concat(raw_x_df, axis=0).reset_index(drop=True)

    def _run(self, drop_bps: bool, user_defined: bool):
        timer = SimbaTimer(start=True)
        self.raw_x_df = self._read_files()
        self.raw_bp_df = self.raw_x_df[
            [Unsupervised.VIDEO.value, Unsupervised.FRAME.value] + self.bp_col_names
        ]
        self.raw_y_df = self.raw_x_df[
            [Unsupervised.FRAME.value, Unsupervised.VIDEO.value] + self.clf_cols
        ]
        self.raw_x_df = self.raw_x_df.drop(self.clf_cols, axis=1)
        if drop_bps:
            self.raw_x_df = self.raw_x_df.drop(self.bp_col_names, axis=1)
        if user_defined:
            self.raw_x_df = self.raw_x_df[
                self.feature_lst + [Unsupervised.FRAME.value, Unsupervised.VIDEO.value]
            ]
        self.feature_names = self.raw_x_df.columns[2:]
        timer.stop_timer()
        stdout_success(
            msg="Reading data complete!", elapsed_time=timer.elapsed_time_str
        )
        self._clf_slicer()
        self.save()

    def _clf_slicer(self):
        bout_data = pd.concat([self.raw_x_df, self.raw_y_df], axis=1)
        bout_data = bout_data.loc[:, ~bout_data.columns.duplicated()].copy()
        self.bouts_x_df = bout_aggregator(
            data=bout_data,
            clfs=self.clf_names,
            video_info=self.video_info_df,
            feature_names=self.feature_names,
            min_bout_length=int(self.settings[Unsupervised.MIN_BOUT_LENGTH.value]),
            aggregator=self.settings[Unsupervised.BOUT_AGGREGATION_TYPE.value],
        )
        if self.settings[Unsupervised.CLF_SLICE_SELECTION.value] in self.clf_names:
            self.bouts_x_df = self.bouts_x_df[
                self.bouts_x_df[Unsupervised.CLASSIFIER.value]
                == self.settings[Unsupervised.CLF_SLICE_SELECTION.value]
            ].reset_index(drop=True)
        self.bouts_y_df = self.bouts_x_df[
            [
                Unsupervised.VIDEO.value,
                Unsupervised.START_FRAME.value,
                Unsupervised.END_FRAME.value,
                Unsupervised.PROBABILITY.value,
                Unsupervised.CLASSIFIER.value,
            ]
        ]
        self.bouts_x_df = self.bouts_x_df.drop(
            [Unsupervised.PROBABILITY.value, Unsupervised.CLASSIFIER.value], axis=1
        )

    def __aggregate_dataset_stats(self):
        stats = {}
        stats["CREATION DATE"] = self.datetime
        stats["FRAME_COUNT"] = len(self.raw_x_df)
        stats["FEATURE_COUNT"] = len(self.feature_names)
        stats["BOUTS_COUNT"] = len(self.bouts_x_df)
        stats["CLASSIFIER_COUNT"] = len(
            self.bouts_y_df[Unsupervised.CLASSIFIER.value].unique()
        )
        for clf in self.bouts_y_df[Unsupervised.CLASSIFIER.value].unique():
            clf_bout_df = self.bouts_y_df[
                self.bouts_y_df[Unsupervised.CLASSIFIER.value] == clf
            ]
            clf_bout_df["LENGTH"] = (
                clf_bout_df[Unsupervised.END_FRAME.value]
                - clf_bout_df[Unsupervised.START_FRAME.value]
            )
            stats[f"{clf}_BOUT_COUNT"] = len(clf_bout_df)
            stats[f"{clf}_MEAN_BOUT_LENGTH (FRAMES)"] = clf_bout_df["LENGTH"].mean()
        stats_df = pd.DataFrame.from_dict(stats, orient="index", columns=["VALUE"])
        stats_df.to_csv(self.log_save_path)
        stdout_success(
            msg=f"Log for unsupervised learning dataset saved at {self.log_save_path}"
        )

    def save(self):
        print("Saving SimBA unsupervised dataset...")
        if len(self.bouts_x_df) == 0:
            raise NoDataError(
                msg="The data contains zero frames after the selected slice setting",
                source=self.__class__.__name__,
            )
        results = {}
        results["DATETIME"] = self.datetime
        results[Unsupervised.BOUT_AGGREGATION_TYPE.name] = self.settings[
            Unsupervised.BOUT_AGGREGATION_TYPE.value
        ]
        results[Unsupervised.MIN_BOUT_LENGTH.name] = self.settings[
            Unsupervised.MIN_BOUT_LENGTH.value
        ]
        results["FEATURE_NAMES"] = self.feature_names
        results["FRAME_FEATURES"] = self.raw_x_df.set_index(
            [Unsupervised.VIDEO.value, Unsupervised.FRAME.value]
        ).astype(np.float32)
        results["FRAME_POSE"] = self.raw_bp_df.set_index(
            [Unsupervised.VIDEO.value, Unsupervised.FRAME.value]
        ).astype(np.int64)
        results["FRAME_TARGETS"] = self.raw_y_df.set_index(
            [Unsupervised.VIDEO.value, Unsupervised.FRAME.value]
        ).astype(np.int8)
        results["BOUTS_FEATURES"] = self.bouts_x_df.set_index(
            [
                Unsupervised.VIDEO.value,
                Unsupervised.START_FRAME.value,
                Unsupervised.END_FRAME.value,
            ]
        ).astype(np.float32)
        results["BOUTS_TARGETS"] = self.bouts_y_df
        write_pickle(data=results, save_path=self.save_path)
        self.timer.stop_timer()
        self.__aggregate_dataset_stats()
        stdout_success(
            msg=f"Dataset for unsupervised learning saved at {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )


# settings = {"data_slice": "ALL FEATURES (EXCLUDING POSE)", "clf_slice": "Attack", "bout_aggregation_type": "MEDIAN", "min_bout_length": 66, "feature_path": "/Users/simon/Desktop/envs/simba_dev/simba/assets/unsupervised/features.csv",}
# db_creator = DatasetCreator(config_path="/Users/simon/Desktop/envs/simba/troubleshooting/unsupervised/project_folder/project_config.ini", settings=settings)
# db_creator.run()


# settings = {"data_slice": "ALL FEATURES (EXCLUDING POSE)", "clf_slice": "ALL CLASSIFIERS (6)", "bout_aggregation_type": "MEDIAN", "min_bout_length": 66, "feature_path": "/Users/simon/Desktop/envs/simba_dev/simba/assets/unsupervised/features.csv",}
# db_creator = DatasetCreator(config_path="/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini", settings=settings)
# db_creator.run()



import os
import time
from copy import deepcopy
from typing import Dict, Optional, Union

import numpy as np
from joblib import Parallel, delayed
from sklearn import clone
from sklearn.ensemble import RandomForestClassifier

try:
    from cuml.ensemble import RandomForestClassifier as cuRF
except:
    from sklearn.ensemble import RandomForestClassifier as cuRF

from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists, check_int,
                                check_valid_array, check_valid_boolean,
                                check_valid_dict)
from simba.utils.enums import Formats
from simba.utils.errors import InvalidInputError
from simba.utils.read_write import find_core_cnt, read_pickle, write_pickle

ACCEPTED_MODELS = RandomForestClassifier, cuRF

class OrdinalClassifier():
    """
    This class implements a strategy for ordinal classification by fitting multiple binary classifiers to predict thresholds between classes.

    It is particularly useful for problems where the target variable has an inherent order but uneven intervals between levels. Thi includes human severity scores, for example, seizures, stereotopy, convulsion, bizarre behavior scores ranging fro 0-5.


    .. warning::
       If larger data sizes (>2m) pass a GPU ``cuml.ensemble.RandomForestClassifier`` object.

    .. note::
       `Modified from sklego <`https://github.com/koaning/scikit-lego/blob/main/sklego/meta/ordinal_classification.py>`__.

    .. csv-table::
       :header: EXPECTED RUNTIMES
       :file: ../../../docs/tables/simba.model.ordinal_clf.OrdinalClassifier.csv
       :widths: 10, 90
       :align: center
       :header-rows: 1

    References
    ----------
    .. [1] Frank, Eibe, and Mark Hall. “A Simple Approach to Ordinal Classification.” In Machine Learning: ECML 2001, edited by Luc De Raedt and Peter Flach, 2167:145–56. Lecture Notes in Computer Science. Berlin, Heidelberg: Springer Berlin Heidelberg, 2001. https://doi.org/10.1007/3-540-44795-4_13.
    .. [2] Sabnis, Gautam, Leinani Hession, J. Matthew Mahoney, Arie Mobley, Marina Santos, and Vivek Kumar. “Visual Detection of Seizures in Mice Using Supervised Machine Learning,” May 31, 2024. https://doi.org/10.1101/2024.05.29.596520.
    .. [3] Another implementation / benchmarking by Lee Prevost - `https://github.com/leeprevost/OrdinalClassifier/tree/main <https://github.com/leeprevost/OrdinalClassifier/tree/main>`_.
    .. [4] https://github.com/fabianp/mord
    .. [5] Michael J Wurm, Paul J Rathouz, and Bret M Hanlon. Regularized ordinal regression and the ordinalnet r package. Journal of Statistical Software, 99(6), 2021.

    :example:
    >>> X = np.random.randint(0, 500, (100, 50))
    >>> y = np.random.randint(1, 6, (100))
    >>> rf_mdl = TrainModelMixin().clf_define(cuda=False)
    >>> fitted_mdl = OrdinalClassifier.fit(X, y, rf_mdl, -1)
    >>> y_hat = OrdinalClassifier.predict_proba(X, fitted_mdl)
    >>> y = OrdinalClassifier.predict(X, fitted_mdl)
    >>> save_path = r"/mnt/c/Users/sroni/Downloads/Box4-20191208T1639-1652/ord_mdl/mdl.pickle"
    >>> OrdinalClassifier.save(mdl=fitted_mdl, save_path=save_path)
    >>> rf_mdl = OrdinalClassifier.load(file_path=save_path)
    >>> y_hat = OrdinalClassifier.predict_proba(X, rf_mdl)
    """

    def __init__(self):
        pass

    @staticmethod
    def fit(X: np.ndarray,
            y: np.ndarray,
            clf: Union[ACCEPTED_MODELS],
            core_cnt: int = -1,
            parallel: Optional[bool] = True) -> Dict[int, Union[ACCEPTED_MODELS]]:

        def _fit_binary_estimator(clf: Union[RandomForestClassifier, cuRF],
                                  X: np.ndarray,
                                  y: np.ndarray,
                                  y_label: int):
            #print(f'Fitting model for y label {y_label}...')
            y_bin = (y <= y_label).astype(np.int32)
            return clone(clf).fit(X, y_bin)

        check_valid_boolean(value=parallel, raise_error=True, source=f'{OrdinalClassifier.__name__} fit parallel')
        classes_ = np.sort(np.unique(y))
        check_valid_array(data=classes_, source=f'{__class__.__name__} y', accepted_ndims=(1,), accepted_dtypes=Formats.INTEGER_DTYPES.value)
        if len(classes_) < 3:
            raise InvalidInputError(msg=f'Found {len(classes_)} classes in y [{classes_}], requires at least 3', source=f'{OrdinalClassifier.__name__} fit')
        intervals = [classes_[i] - classes_[i-1] for i in range(1, len(classes_))]
        if len(set(intervals)) != 1:
            raise InvalidInputError(msg=f'The values in y ({classes_}) are not of equal interval.', source=f'{OrdinalClassifier.__name__} fit')
        check_valid_array(data=X, source=f'{__class__.__name__} x', accepted_ndims=(2,), accepted_dtypes=Formats.NUMERIC_DTYPES.value, accepted_axis_0_shape=(y.shape[0],))
        if not isinstance(clf, (RandomForestClassifier, cuRF)) or ('predict_proba' not in dir(clf)):
            raise InvalidInputError(msg=f'clf is not of valid type: {type(clf)} (accepted: {ACCEPTED_MODELS})', source=f'{OrdinalClassifier.__name__} fit')
        parallel = False if isinstance(clf, cuRF) else parallel
        check_int(name='core_cnt', min_value=-1, unaccepted_vals=[0], value=core_cnt)
        core_cnt = [find_core_cnt()[0] if core_cnt == -1 or core_cnt > find_core_cnt()[0] else core_cnt][0]
        if parallel:
            return dict(zip(classes_[:-1], Parallel(n_jobs=core_cnt)(delayed(_fit_binary_estimator)(clf, X, y, y_label) for y_label in classes_[:-1])))
        else:
            results = {}
            for y_label in classes_[:-1]:
                results[y_label] = _fit_binary_estimator(clf=clf, X=X, y=y, y_label=y_label)
            return results


    @staticmethod
    def predict_proba(X: np.ndarray, mdl: Dict[int, Union[ACCEPTED_MODELS]]) -> np.ndarray:
        OrdinalClassifier._check_valid_mdl_dict(mdls=mdl)
        check_valid_array(data=X, source=f'{__class__.__name__} x', accepted_ndims=(2,), accepted_dtypes=Formats.NUMERIC_DTYPES.value)
        n_features = mdl[list(mdl.keys())[0]].n_features_ if hasattr(mdl, 'n_features_') else mdl[list(mdl.keys())[0]].n_features_in_

        if n_features != X.shape[1]:
            raise InvalidInputError(msg=f'Model expects {mdl[list(mdl.keys())[0]].n_features_} features, got {X.shape[1]}.', source=f'{OrdinalClassifier.__name__} predict')
        check_valid_array(data=X, source=f'{__class__.__name__} x', accepted_ndims=(2,), accepted_dtypes=Formats.NUMERIC_DTYPES.value)
        raw_proba = np.array([estimator.predict_proba(X)[:, 1] for estimator in mdl.values()]).T
        return np.diff(np.column_stack((np.zeros(X.shape[0]), raw_proba, np.ones(X.shape[0]))), n=1, axis=1)


    @staticmethod
    def predict(X: np.ndarray, mdl: Dict[int, Union[ACCEPTED_MODELS]]) -> np.ndarray:
        OrdinalClassifier._check_valid_mdl_dict(mdls=mdl)
        check_valid_array(data=X, source=f'{__class__.__name__} x', accepted_ndims=(2,), accepted_dtypes=Formats.NUMERIC_DTYPES.value)
        n_features = mdl[list(mdl.keys())[0]].n_features_ if hasattr(mdl, 'n_features_') else mdl[list(mdl.keys())[0]].n_features_in_

        if n_features != X.shape[1]:
            raise InvalidInputError(msg=f'Model expects {mdl[list(mdl.keys())[0]].n_features_} features, got {X.shape[1]}.', source=f'{OrdinalClassifier.__name__} predict')
        return np.argmax(OrdinalClassifier.predict_proba(X, mdl=mdl), axis=1)

    @staticmethod
    def save(mdl: Dict[int, Union[ACCEPTED_MODELS]], save_path: Union[str, os.PathLike]):
        OrdinalClassifier._check_valid_mdl_dict(mdls=mdl)
        check_if_dir_exists(in_dir=os.path.dirname(save_path), source=f'{OrdinalClassifier.__name__} save')
        write_pickle(data=mdl, save_path=save_path)


    @staticmethod
    def load(file_path: Union[str, os.PathLike]) -> Dict[int, Union[ACCEPTED_MODELS]]:
        check_file_exist_and_readable(file_path=file_path)
        return read_pickle(data_path=file_path)


    @staticmethod
    def _check_valid_mdl_dict(mdls: Dict[int, Union[ACCEPTED_MODELS]]) -> None:
        check_valid_dict(x=mdls, valid_keys=list(range(0, 5000)), min_len_keys=2, valid_key_dtypes=Formats.INTEGER_DTYPES.value, valid_values_dtypes=ACCEPTED_MODELS)
        features_in_cnt = []
        for mdl in mdls.values():
            features_in_cnt.append(mdl.n_features_) if hasattr(mdl, 'n_features_') else features_in_cnt.append(mdl.n_features_in_)
        if len(set(features_in_cnt)) != 1:
            raise InvalidInputError(msg=f'The models has different N features [{features_in_cnt}]')




#
# for
#


# NS = [1000, 10000, 100000, 1000000, 2000000]
# CUDA = False
# for N in NS:
#     X = np.random.randint(0, 500, (N, 250)).astype(np.int32)
#     y = np.random.randint(1, 6, (N)).astype(np.int32)
#     rf_mdl = TrainModelMixin().clf_define(cuda=CUDA, verbose=False, n_estimators=250)
#     start = time.perf_counter()
#     fitted_mdl = OrdinalClassifier.fit(X, y, rf_mdl, -1, parallel=False)
#     stop = time.perf_counter()
#     elapsed = stop - start
#     print(N, '\t'* 4, elapsed)

#
#
# y_hat = OrdinalClassifier.predict_proba(X, fitted_mdl)
# y = OrdinalClassifier.predict(X, fitted_mdl)
# save_path = r"/mnt/c/Users/sroni/Downloads/Box4-20191208T1639-1652/ord_mdl/mdl.pickle"
# OrdinalClassifier.save(mdl=fitted_mdl, save_path=save_path)
# rf_mdl = OrdinalClassifier.load(file_path=save_path)
# y_hat = OrdinalClassifier.predict_proba(X, rf_mdl)
#

# X = np.random.randint(0, 500, (100, 50))
# y = np.random.randint(1, 6, (100))
# rf_mdl = TrainModelMixin().clf_define()
# fitted_mdls = OrdinalClassifier.fit(X, y, rf_mdl, -1)
# y_hat = OrdinalClassifier.predict_proba(X, fitted_mdls)
# y = OrdinalClassifier.predict(X, fitted_mdls)
# OrdinalClassifier.save(mdl=fitted_mdls, save_path=r"C:\Users\sroni\OneDrive\Desktop\mdl.pk")

#predict_proba(X)
# ordinal_clf.predict(X)


__author__ = "Simon Nilsson"

import os
import tkinter.ttk as ttk
from copy import deepcopy
from tkinter import *

import pandas as pd
import PIL.Image
from PIL import ImageTk

import simba
from simba.ui.import_pose_frame import ImportPoseFrame
from simba.ui.import_videos_frame import ImportVideosFrame
from simba.ui.pop_ups.clf_add_remove_print_pop_up import PoseResetterPopUp
from simba.ui.pop_ups.create_user_defined_pose_configuration_pop_up import \
    CreateUserDefinedPoseConfigurationPopUp
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, Entry_Box,
                                        FolderSelect, SimbaButton,
                                        SimBADropDown, hxtScrollbar)
from simba.utils.checks import check_if_dir_exists, check_str, check_int
from simba.utils.config_creator import ProjectConfigCreator
from simba.utils.enums import Formats, Keys, Links, Methods, Options, Paths
from simba.utils.errors import DuplicationError, MissingProjectConfigEntryError
from simba.utils.lookups import (get_body_part_configurations,
                                 get_bp_config_codes, get_icons_paths)
from simba.video_processors.video_processing import \
    extract_frames_from_all_videos_in_directory


class ProjectCreatorPopUp():
    """
    Mixin for GUI pop-up windows that accept user-inputs for creating a SimBA project.

    .. image:: _static/img/ProjectCreatorPopUp.webp
       :width: 800
       :align: center

    :example:
    >>> ProjectCreatorPopUp()
    """

    def __init__(self):

        self.main_frm = Toplevel()
        self.main_frm.minsize(750, 800)
        self.main_frm.wm_title("PROJECT CONFIGURATION")
        self.main_frm.columnconfigure(0, weight=1)
        self.main_frm.rowconfigure(0, weight=1)
        parent_tab = ttk.Notebook(hxtScrollbar(self.main_frm))
        self.btn_icons = get_icons_paths()
        for k in self.btn_icons.keys():
            self.btn_icons[k]["img"] = ImageTk.PhotoImage(image=PIL.Image.open(os.path.join(os.path.dirname("__file__"), self.btn_icons[k]["icon_path"])))
        self.main_frm.iconphoto(False, self.btn_icons['settings']['img'])

        self.create_project_tab = ttk.Frame(parent_tab)
        self.import_videos_tab = ttk.Frame(parent_tab)
        self.import_data_tab = ttk.Frame(parent_tab)
        #.extract_frms_tab = ttk.Frame(parent_tab)

        parent_tab.add(self.create_project_tab, text=f'{"[ Create project config ]": ^20s}', compound="left", image=self.btn_icons["create"]["img"])
        parent_tab.add(self.import_videos_tab, text=f'{"[ Import videos ]": ^20s}', compound="left", image=self.btn_icons["video"]["img"])
        parent_tab.add(self.import_data_tab, text=f'{"[ Import tracking data ]": ^20s}', compound="left", image=self.btn_icons["pose"]["img"])
        #parent_tab.add( self.extract_frms_tab, text=f'{"[ Extract frames ]": ^20s}', compound="left", image=self.btn_icons["frames"]["img"])
        parent_tab.grid(row=0, column=0, sticky=NW)

        self.settings_frm = CreateLabelFrameWithIcon(parent=self.create_project_tab, header="SETTINGS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.CREATE_PROJECT.value)
        self.general_settings_frm = CreateLabelFrameWithIcon(parent=self.settings_frm, header="GENERAL PROJECT SETTINGS", icon_name='settings', icon_link=Links.CREATE_PROJECT.value, padx=5, pady=5, relief='solid')
        self.project_dir_select = FolderSelect(self.general_settings_frm, "PROJECT DIRECTORY:", lblwidth=35, entry_width=35, font=Formats.FONT_REGULAR.value)
        self.project_name_eb = Entry_Box(self.general_settings_frm, "PROJECT NAME:", labelwidth=35, entry_box_width=35)
        self.file_type_dropdown = SimBADropDown(parent=self.general_settings_frm, dropdown_options=Options.WORKFLOW_FILE_TYPE_OPTIONS.value, label='WORKFLOW FILE TYPE:', label_width=35, dropdown_width=35, value=Options.WORKFLOW_FILE_TYPE_OPTIONS.value[0])

        self.clf_entry_boxes = []
        self.ml_settings_frm = CreateLabelFrameWithIcon(parent=self.create_project_tab, header="MACHINE LEARNING SETTINGS", icon_name='forest', icon_link=Links.CREATE_PROJECT.value, font=Formats.FONT_HEADER.value, padx=5, pady=5, relief='solid')
        self.clf_cnt_dropdown = SimBADropDown(parent=self.ml_settings_frm, dropdown_options=list(range(1, 26)), label='NUMBER OF CLASSIFIERS (BEHAVIORS)', label_width=35, dropdown_width=35, value=1, command=self.__create_entry_boxes)
        self.__create_entry_boxes(cnt=1)
        self.animal_settings_frm = CreateLabelFrameWithIcon(parent=self.create_project_tab, header="ANIMAL SETTINGS", icon_name='pose', icon_link=Links.CREATE_PROJECT.value, font=Formats.FONT_HEADER.value, padx=5, pady=5, relief='solid')
        self.tracking_type_dropdown = SimBADropDown(parent=self.animal_settings_frm, dropdown_options=Options.TRACKING_TYPE_OPTIONS.value, label='TYPE OF TRACKING', label_width=35, dropdown_width=35, value=Options.TRACKING_TYPE_OPTIONS.value[0], command=self.update_body_part_dropdown)

        project_animal_cnt_path = os.path.join(os.path.dirname(simba.__file__), Paths.SIMBA_NO_ANIMALS_PATH.value)
        self.animal_count_lst = list(pd.read_csv(project_animal_cnt_path, header=None)[0])
        self.bp_lu = get_body_part_configurations()
        self.bp_config_codes = get_bp_config_codes()
        self.classical_tracking_options = deepcopy(Options.CLASSICAL_TRACKING_OPTIONS.value)
        self.multi_tracking_options = deepcopy(Options.MULTI_ANIMAL_TRACKING_OPTIONS.value)
        self.three_dim_tracking_options = deepcopy(Options.THREE_DIM_TRACKING_OPTIONS.value)
        self.user_defined_options = [x
                                     for x in list(self.bp_lu.keys())
                                     if x not in self.classical_tracking_options
                                     and x not in self.multi_tracking_options
                                     and x not in self.three_dim_tracking_options]
        for k in self.bp_lu.keys():
            self.bp_lu[k]["img"] = ImageTk.PhotoImage(file=os.path.join(os.path.dirname("__file__"), self.bp_lu[k]["img_path"]))

        self.classical_tracking_option_dict = {k: self.bp_lu[k] for k in self.classical_tracking_options}
        self.multi_tracking_option_dict = {k: self.bp_lu[k] for k in self.multi_tracking_options}
        self.classical_tracking_options.append(Methods.CREATE_POSE_CONFIG.value)
        self.multi_tracking_options.append(Methods.CREATE_POSE_CONFIG.value)
        self.three_dim_tracking_options.append(Methods.CREATE_POSE_CONFIG.value)
        self.classical_tracking_options.extend(self.user_defined_options)
        self.multi_tracking_options.extend(self.user_defined_options)
        self.three_dim_tracking_options.extend(self.user_defined_options)

        self.selected_tracking_dropdown = SimBADropDown(parent=self.animal_settings_frm, dropdown_options=Options.CLASSICAL_TRACKING_OPTIONS.value, label='BODY-PART CONFIGURATION: ', label_width=35, dropdown_width=35, value=self.classical_tracking_options[0], command=self.update_img)
        self.img_lbl = Label(self.animal_settings_frm, image=self.bp_lu[self.classical_tracking_options[0]]["img"], font=Formats.FONT_REGULAR.value)
        reset_btn = SimbaButton(parent=self.animal_settings_frm, txt="RESET USER DEFINED POSE-CONFIGS", txt_clr='red', img='clean', cmd=PoseResetterPopUp)

        run_frm = CreateLabelFrameWithIcon(parent=self.create_project_tab, header="CREATE PROJECT CONFIG", icon_name='create', icon_link=Links.CREATE_PROJECT.value)
        create_project_btn = SimbaButton(parent=run_frm, txt="CREATE PROJECT CONFIG", txt_clr='navy', img='create', font=Formats.FONT_HEADER.value, cmd=self.run)

        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.general_settings_frm.grid(row=0, column=0, sticky=NW, pady=5)
        self.project_dir_select.grid(row=0, column=0, sticky=NW)
        self.project_name_eb.grid(row=1, column=0, sticky=NW)
        self.file_type_dropdown.grid(row=2, column=0, sticky=NW)

        self.ml_settings_frm.grid(row=1, column=0, sticky=NW, pady=5)
        self.clf_cnt_dropdown.grid(row=0, column=0, sticky=NW)

        self.animal_settings_frm.grid(row=2, column=0, sticky=NW, pady=5)
        self.tracking_type_dropdown.grid(row=0, column=0, sticky=NW)
        self.selected_tracking_dropdown.grid(row=1, column=0, sticky=NW)
        self.img_lbl.grid(row=2, column=0, sticky=NW)
        reset_btn.grid(row=0, column=1, sticky=NW)
        run_frm.grid(row=3, column=0, sticky=NW)
        create_project_btn.grid(row=0, column=0, sticky=NW)

        ImportVideosFrame(parent_frm=self.import_videos_tab, config_path=None, idx_row=0, idx_column=0)
        ImportPoseFrame(parent_frm=self.import_data_tab, config_path=None, idx_row=0, idx_column=0)
        #extract_frames_frm = LabelFrame(self.extract_frms_tab, text="EXTRACT FRAMES INTO PROJECT", fg="black", font=Formats.FONT_HEADER.value, pady=5, padx=5)
        #extract_frames_note = Label(extract_frames_frm, text="Note: Frame extraction is not needed for any of the parts of the SimBA pipeline.\n Caution: This extract all frames from all videos in project. \n and is computationally expensive if there is a lot of videos at high frame rates/resolution.", font=Formats.FONT_REGULAR.value)
        #extract_frames_btn = SimbaButton(parent=extract_frames_frm, txt="EXTRACT FRAMES", txt_clr='blue', font=Formats.FONT_REGULAR.value, cmd=None)
        #extract_frames_frm.grid(row=0, column=0, sticky=NW)
        #extract_frames_note.grid(row=0, column=0, sticky=NW)
        #extract_frames_btn.grid(row=1, column=0, sticky=NW)
        self.update_body_part_dropdown(Methods.CLASSIC_TRACKING.value)
        #self.main_frm.mainloop()

    def __create_entry_boxes(self, cnt):
        existing_values = []
        for entry in self.clf_entry_boxes:
            try:
                existing_values.append(entry.entry_get.strip())
            except:
                existing_values.append("")
        for entry in self.clf_entry_boxes:
            try:
                entry.destroy()
            except:
                pass
        self.clf_entry_boxes = []
        valid_cnt, _ = check_int(name=f'{self.__class__.__name__} cnt', value=cnt, min_value=1, raise_error=False)
        count = int(cnt) if valid_cnt else 1
        for clf_cnt in range(count):
            entry = Entry_Box(parent=self.ml_settings_frm, fileDescription=f'CLASSIFIER NAME {clf_cnt + 1}: ', labelwidth=35, entry_box_width=35)
            entry.grid(row=clf_cnt + 2, column=0, sticky=NW)
            if clf_cnt < len(existing_values) and existing_values[clf_cnt]:
                try:
                    entry.entry_set(existing_values[clf_cnt])
                except AttributeError:
                    try:
                        entry.entPath.insert(0, existing_values[clf_cnt])
                    except:
                        pass
            self.clf_entry_boxes.append(entry)
        # for entry in self.clf_entry_boxes:
        #     entry.destroy()
        # self.clf_entry_boxes = []
        # for clf_cnt in range(int(cnt)):
        #     entry = Entry_Box(parent=self.ml_settings_frm, fileDescription=f'CLASSIFIER NAME {clf_cnt + 1}: ', labelwidth=35, entry_box_width=35)
        #     entry.grid(row=clf_cnt + 2, column=0, sticky=NW)
        #     self.clf_entry_boxes.append(entry)

    def update_body_part_dropdown(self, selected_value):
        self.selected_tracking_dropdown.destroy()
        if selected_value == Methods.MULTI_TRACKING.value:
            self.selected_tracking_dropdown = SimBADropDown(parent=self.animal_settings_frm, dropdown_options=self.multi_tracking_options, label='BODY-PART CONFIGURATION: ', label_width=35, dropdown_width=35, value=self.multi_tracking_options[0], command=self.update_img)
            self.selected_tracking_dropdown.grid(row=1, column=0, sticky=NW)
        elif selected_value == Methods.CLASSIC_TRACKING.value:
            self.selected_tracking_dropdown = SimBADropDown(parent=self.animal_settings_frm, dropdown_options=self.classical_tracking_options, label='BODY-PART CONFIGURATION: ', label_width=35, dropdown_width=35, value=self.classical_tracking_options[0], command=self.update_img)
            self.selected_tracking_dropdown.grid(row=1, column=0, sticky=NW)

        elif selected_value == Methods.THREE_D_TRACKING.value:
            self.selected_tracking_dropdown = SimBADropDown(parent=self.animal_settings_frm, dropdown_options=self.three_dim_tracking_options, label='BODY-PART CONFIGURATION: ', label_width=35, dropdown_width=35, value=self.three_dim_tracking_options[0], command=self.update_img)
            self.selected_tracking_dropdown.grid(row=1, column=0, sticky=NW)
        self.update_img(self.selected_tracking_dropdown.getChoices())

    def update_img(self, selected_value):
        if selected_value != Methods.CREATE_POSE_CONFIG.value:
            self.img_lbl.config(image=self.bp_lu[selected_value]["img"])
        else:
            _ = CreateUserDefinedPoseConfigurationPopUp(master=self.main_frm, project_config_class=ProjectCreatorPopUp)

    def extract_frames(self):
        if not hasattr(self, "config_path"):
            raise MissingProjectConfigEntryError(msg="Create PROJECT CONFIG before extracting frames")
        video_dir = os.path.join(os.path.dirname(self.config_path), "videos")
        extract_frames_from_all_videos_in_directory(config_path=self.config_path, directory=video_dir)

    def run(self):
        project_dir = self.project_dir_select.folder_path
        check_if_dir_exists(in_dir=project_dir)
        project_name = self.project_name_eb.entry_get
        check_str(name="PROJECT NAME", value=project_name, allow_blank=False)
        target_list = []
        for number, entry_box in enumerate(self.clf_entry_boxes):
            target_list.append(entry_box.entry_get.strip())
        if len(list(set(target_list))) != len(self.clf_entry_boxes):
            raise DuplicationError(msg="All classifier names have to be unique")
        selected_config = self.selected_tracking_dropdown.getChoices()
        if selected_config in self.bp_config_codes.keys():
            config_code = self.bp_config_codes[selected_config]
        else:
            config_code = Methods.USER_DEFINED.value

        config_idx = None
        for cnt, k in enumerate(self.bp_lu.keys()):
            if k == selected_config:
                config_idx = cnt
        animal_cnt = self.animal_count_lst[config_idx]

        config_creator = ProjectConfigCreator(project_path=project_dir,
                                              project_name=project_name,
                                              target_list=target_list,
                                              pose_estimation_bp_cnt=config_code,
                                              body_part_config_idx=config_idx,
                                              animal_cnt=animal_cnt,
                                              file_type=self.file_type_dropdown.getChoices())

        self.config_path = config_creator.config_path
        ImportPoseFrame(parent_frm=self.import_data_tab, idx_row=0, idx_column=0, config_path=self.config_path)
        ImportVideosFrame(parent_frm=self.import_videos_tab, config_path=self.config_path, idx_row=0, idx_column=0)

#ProjectCreatorPopUp()



__author__ = "Simon Nilsson"

import os
import threading
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.roi_tools.ROI_feature_analyzer import ROIFeatureCreator
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        SimbaButton)
from simba.utils.enums import Formats, Keys, Links
from simba.utils.warnings import ROIWarning


class AppendROIFeaturesByAnimalPopUp(ConfigReader, PopUpMixin):
    def __init__(self, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        if not os.path.isfile(self.roi_coordinates_path):
            ROIWarning(msg=f"SIMBA ERROR: No ROIs have been defined. Please define ROIs before appending ROI-based features (no data file found at path {self.roi_coordinates_path})", source=self.__class__.__name__,)

        PopUpMixin.__init__(self, title="APPEND ROI FEATURES: BY ANIMALS", size=(400, 400), icon='shapes_small')
        self.animal_cnt_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SELECT NUMBER OF ANIMALS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.ROI_FEATURES.value)
        self.animal_cnt_dropdown = DropDownMenu(self.animal_cnt_frm, "# of animals", list(range(1, self.animal_cnt + 1)), labelwidth=20)
        self.animal_cnt_dropdown.setChoices(1)
        self.animal_cnt_confirm_btn = SimbaButton(parent=self.animal_cnt_frm, txt='Confirm', img='tick', cmd=self.create_settings_frm, height=25)

        self.animal_cnt_frm.grid(row=0, column=0, sticky=NW)
        self.animal_cnt_dropdown.grid(row=0, column=0, sticky=NW)
        self.animal_cnt_confirm_btn.grid(row=0, column=1, sticky=NW)
        self.main_frm.mainloop()

    def create_settings_frm(self):
        if hasattr(self, "setting_frm"):
            self.setting_frm.destroy()
            self.body_part_frm.destroy()
        self.setting_frm = LabelFrame(
            self.main_frm, text="SETTINGS", font=Formats.FONT_HEADER.value
        )
        self.choose_bp_frm(parent=self.setting_frm, bp_options=self.body_parts_lst)
        self.setting_frm.grid(row=1, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

    def run(self):
        body_parts = []
        for bp_cnt, bp_dropdown in self.body_parts_dropdowns.items():
            body_parts.append(bp_dropdown.getChoices())

        roi_feature_creator = ROIFeatureCreator(
            config_path=self.config_path,
            body_parts=body_parts,
            data_path=None,
            append_data=True,
        )
        threading.Thread(target=roi_feature_creator.run()).start()
        self.root.destroy()


#AppendROIFeaturesByAnimalPopUp(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")


# AppendROIFeaturesByAnimalPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/locomotion/project_folder/project_config.ini')


# AppendROIFeaturesByAnimalPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



from simba.utils.data import get_library_version
from simba.utils.enums import ENV_VARS
from simba.utils.read_write import read_sys_env
from simba.utils.warnings import VersionWarning

ENV = read_sys_env()

### IMPORT CURF IF SET BY simba/assets/.env
use_cuml = ENV.get(ENV_VARS.CUML.value, False)  # Avoid KeyError, default to False
#use_cuml = True
if use_cuml:
    cuml_version = get_library_version(library_name='cuml', raise_error=False)
    if cuml_version:
        try:
            from cuml.ensemble import RandomForestClassifier as cuRF
            print('SimBA CUML enabled.')
        except ImportError:
            VersionWarning(msg="cuML is set but not installed. Falling back to scikit-learn.")
            from sklearn.ensemble import RandomForestClassifier as cuRF
    else:
        print("cuML is set but not installed. Falling back to scikit-learn.")
        from sklearn.ensemble import RandomForestClassifier as cuRF
else:
    from sklearn.ensemble import RandomForestClassifier as cuRF

__all__ = ["cuRF"]


import itertools
import os
from typing import Dict, List, Optional, Tuple, Union

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import networkx as nx
import numpy as np
from numba import jit
from pyvis.network import Network

from simba.utils.checks import (check_float, check_if_dir_exists,
                                check_instance, check_int,
                                check_iterable_length, check_str,
                                check_valid_array, check_valid_hex_color,
                                check_valid_lst, check_valid_tuple)
from simba.utils.data import create_color_palette, find_ranked_colors, get_mode
from simba.utils.enums import Formats
from simba.utils.errors import CountError, InvalidInputError


class NetworkMixin(object):
    """
    Methods to create and analyze time-dependent graphs from pose-estimation data.

    When working with pose-estimation data for more than two animals - over extended periods - it can be beneficial to
    represent the data as a graph where the animals feature as nodes are their relationship strengths are represented as edges.

    When formatted as a graph, we can compute (i) how the relationships between animal pairs change across time and recordings,
    (ii) the relative importance's and hierarchies of individual animals within the group, or (iii) identify sub-groups with the network.

    The critical component determining the results is how edge weights are represented. These edge weight values could be the amount of time animal bounding
    boxes overlap each other, aggregate distances between the animals, or how much time animals engange in coordinated behaviors. These values can be computed
    through other methods within SimBA mixin methods.

    Very much wip and so far primarily depend on `networkx <https://github.com/networkx/networkx>`_.

    References
    ----------
    See below references for mature and reliable packages (12/2023):

    .. [1] `networkx <https://github.com/networkx/networkx>`_
    .. [2] `igraph <https://igraph.org/>`_
    """

    def __init__(self):
        pass

    @staticmethod
    def create_graph(data: Dict[Tuple[str, str], float]) -> nx.Graph():
        """
        Create a single undirected graph with single edges from on dictionary.

        :param Dict[Tuple[str, str], float] data: A dictionary where keys are tuples representing node pairs and values are the corresponding edge weights.
        :returns nx.Graph: A networkx graph with nodes and edges defined by the input data.

        :example:
        >>> data = {('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5}
        >>> graph = NetworkMixin.create_graph(data=data)
        """

        check_instance(source=NetworkMixin.create_graph.__name__, instance=data, accepted_types=dict)
        for k, v in data.items():
            check_instance(source=NetworkMixin.create_graph.__name__, instance=k,accepted_types=tuple)
            check_iterable_length(source=f"{NetworkMixin.create_graph.__name__} {k}", val=len(k), exact_accepted_length=2)
            check_instance(source=f"{NetworkMixin.create_graph.__name__} {v}", instance=v, accepted_types=Formats.NUMERIC_DTYPES.value)

        G = nx.Graph()
        for node_names in data.keys():
            G.add_node(node_names[0])
            G.add_node(node_names[1])
        for node_names, edge_weight in data.items():
            G.add_edge(node_names[0], node_names[1], weight=edge_weight)
        return G

    @staticmethod
    def create_multigraph(data: Dict[Tuple[str, str], List[float]]) -> nx.MultiGraph:
        """
        Create a multi-graph from a dictionary of node pairs and associated edge weights.

        For example, creates a multi-graph where node edges represent animal relationship weights at different
        timepoints.

        :param Dict[Tuple[str, str], List[float]] data: A dictionary where keys are tuples representing node pairs, and values are lists of edge weights associated with each pair.
        :returns nx.MultiGraph: A NetworkX multigraph with nodes and edges specified by the input data. Each edge is labeled and weighted based on the provided information.

        :example:
        >>> data = {('Animal_1', 'Animal_2'): [0, 0, 0, 6], ('Animal_1', 'Animal_3'): [0, 0, 0, 0], ('Animal_1', 'Animal_4'): [0, 0, 0, 0], ('Animal_1', 'Animal_5'): [0, 0, 0, 0], ('Animal_2', 'Animal_3'): [0, 0, 0, 0], ('Animal_2', 'Animal_4'): [5, 0, 0, 2], ('Animal_2', 'Animal_5'): [0, 0, 0, 0], ('Animal_3', 'Animal_4'): [0, 0, 0, 0], ('Animal_3', 'Animal_5'): [0, 2, 22, 0], ('Animal_4', 'Animal_5'): [0, 0, 0, 0]}
        >>> NetworkMixin().create_multigraph(data=data)
        """

        check_instance(source=NetworkMixin.create_multigraph.__name__, instance=data, accepted_types=dict)
        results, dict_data, G = {}, [], nx.MultiGraph()
        for k, v in data.items():
            check_instance(
                source=NetworkMixin.create_multigraph.__name__,
                instance=k,
                accepted_types=tuple,
            )
            check_instance(
                source=f"{NetworkMixin.create_multigraph.__name__} {v}",
                instance=v,
                accepted_types=list,
            )
            dict_data.append(len(v))
        check_iterable_length(
            source=f"{NetworkMixin.create_multigraph.__name__} data",
            val=len(list(set([x for x in dict_data]))),
            exact_accepted_length=1,
        )
        for node_names in data.keys():
            G.add_node(node_names[0])
            G.add_node(node_names[1])
        for node_names, edge_weights in data.items():
            for edge_cnt, edge_weight in enumerate(edge_weights):
                G.add_edge(
                    node_names[0],
                    node_names[1],
                    weight=edge_weight,
                    label=f"edge_{edge_cnt}",
                )
        return G

    @staticmethod
    def graph_page_rank(graph: nx.Graph,
                            weights: Optional[str] = "weight",
                            alpha: Optional[float] = 0.85,
                            max_iter: Optional[int] = 100) -> Dict[str, float]:

        """
        Calculate the PageRank of nodes in a graph.

        :example:
        >>> graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
        >>> NetworkMixin().graph_page_rank(graph=graph)
        """

        check_instance(
            source=NetworkMixin.graph_page_rank.__name__,
            instance=graph,
            accepted_types=nx.Graph,
        )
        check_str(
            name=f"{NetworkMixin.graph_page_rank.__name__} weights", value=weights
        )
        check_float(name=f"{NetworkMixin.graph_page_rank.__name__} alpha", value=alpha)
        check_int(
            name=f"{NetworkMixin.graph_page_rank.__name__} max_iter",
            value=max_iter,
            min_value=1,
        )
        edge_weights = tuple(
            set(
                itertools.chain.from_iterable(
                    d.keys() for *_, d in graph.edges(data=True)
                )
            )
        )
        check_str(
            name=f"{NetworkMixin.graph_page_rank.__name__} weights",
            value=weights,
            options=edge_weights,
        )

        return nx.pagerank(graph, alpha=alpha, max_iter=max_iter, weight=weights)

    @staticmethod
    def graph_katz_centrality(graph: nx.Graph, weights: Optional[str] = "weight", alpha: Optional[float] = 0.85,):
        """
        Katz centrality is an algorithm in NetworkX that measures the relative influence of a node in a network.

        See networkx documentation

        :example:
        >>> graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
        >>> NetworkMixin().graph_katz_centrality(graph=graph)
        """
        check_instance(source=NetworkMixin.graph_katz_centrality.__name__, instance=graph, accepted_types=nx.Graph,)
        check_str(name=f"{NetworkMixin.graph_katz_centrality.__name__} weights", value=weights)
        check_float(name=f"{NetworkMixin.graph_katz_centrality.__name__} alpha", value=alpha)
        edge_weights = tuple(
            set(
                itertools.chain.from_iterable(
                    d.keys() for *_, d in graph.edges(data=True)
                )
            )
        )
        check_str(
            name=f"{NetworkMixin.graph_katz_centrality.__name__} weights",
            value=weights,
            options=edge_weights,
        )

        return nx.katz_centrality_numpy(graph, alpha=alpha, weight=weights)

    @staticmethod
    def graph_current_flow_closeness_centrality(graph: nx.Graph, weights: Optional[str] = "weight"):

        """
        :example:
        >>> graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
        >>> NetworkMixin().graph_current_flow_closeness_centrality(graph=graph)
        """

        check_instance(
            source=NetworkMixin.graph_current_flow_closeness_centrality.__name__,
            instance=graph,
            accepted_types=nx.Graph,
        )
        check_str(
            name=f"{NetworkMixin.graph_current_flow_closeness_centrality.__name__} weights",
            value=weights,
        )
        edge_weights = tuple(
            set(
                itertools.chain.from_iterable(
                    d.keys() for *_, d in graph.edges(data=True)
                )
            )
        )
        check_str(
            name=f"{NetworkMixin.graph_current_flow_closeness_centrality.__name__} weights",
            value=weights,
            options=edge_weights,
        )
        return nx.current_flow_closeness_centrality(
            graph, weight=weights, solver="full"
        )

    @staticmethod
    def girvan_newman(
        graph: nx.Graph,
        levels: Optional[int] = 1,
        most_valuable_edge: Optional[object] = None,
    ):
        """
        :example:
        >>> graph = NetworkMixin.create_graph({ ('Animal_1', 'Animal_2'): 0.0, ('Animal_1', 'Animal_3'): 0.0, ('Animal_1', 'Animal_4'): 0.0, ('Animal_1', 'Animal_5'): 0.0, ('Animal_2', 'Animal_3'): 1.0, ('Animal_2', 'Animal_4'): 1.0, ('Animal_2', 'Animal_5'): 1.0, ('Animal_3', 'Animal_4'): 1.0, ('Animal_3', 'Animal_5'): 1.0, ('Animal_4', 'Animal_5'): 1.0})
        >>> NetworkMixin().girvan_newman(graph=graph, levels = 1)
        >>> [({'Animal_1'}, {'Animal_2', 'Animal_3', 'Animal_4', 'Animal_5'})]
        """

        check_instance(
            source=NetworkMixin.multigraph_page_rank.__name__,
            instance=graph,
            accepted_types=nx.Graph,
        )
        if levels > graph.number_of_nodes():
            raise CountError(
                msg=f"Number of nodes ({G.number_of_nodes()}) is less than numer of girvan newman levels ({levels}).",
                source=NetworkMixin.__class__.__name__,
            )

        communities = list(
            nx.algorithms.community.girvan_newman(
                graph, most_valuable_edge=most_valuable_edge
            )
        )
        return communities[:levels]

    @staticmethod
    def multigraph_page_rank(
        graph: nx.MultiGraph,
        weights: Optional[str] = "weight",
        alpha: Optional[float] = 0.85,
        max_iter: Optional[int] = 100,
    ) -> Dict[str, List[float]]:
        """
        Calculate multi-graph PageRank scores for each node in a MultiGraph.

        For example, each node-pair in a graph has N undirected edges representing the weighted relationship between the two nodes atobserved point in time.
        Calculates the page rank of each node at each observed time point.

        :param nx.MultiGraph graph: The input MultiGraph, created by ``NetworkMixin.create_multigraph()``.

        :example:
        >>> multigraph = NetworkMixin().create_multigraph(data={('Animal_1', 'Animal_2'): [0, 0, 0, 6], ('Animal_1', 'Animal_3'): [0, 0, 0, 0], ('Animal_1', 'Animal_4'): [0, 0, 0, 0], ('Animal_1', 'Animal_5'): [0, 0, 0, 0], ('Animal_2', 'Animal_3'): [0, 0, 0, 0], ('Animal_2', 'Animal_4'): [5, 0, 0, 2], ('Animal_2', 'Animal_5'): [0, 0, 0, 0], ('Animal_3', 'Animal_4'): [0, 0, 0, 0], ('Animal_3', 'Animal_5'): [0, 2, 22, 0], ('Animal_4', 'Animal_5'): [0, 0, 0, 0]})
        >>> NetworkMixin().multigraph_page_rank(graph=multigraph)
        >>> {'Animal_1': [0.06122524589028524, 0.06122524589028524, 0.06122524589028524, 0.32739635847890775], 'Animal_2': [0.06122524589028524, 0.40816213116457223, 0.06122524589028524, 0.442259400816002], 'Animal_3': [0.40816213116457223, 0.06122524589028524, 0.40816213116457223, 0.04545454545454547], 'Animal_4': [0.06122524589028524, 0.40816213116457223, 0.06122524589028524, 0.13943514979599955], 'Animal_5': [0.40816213116457223, 0.06122524589028524, 0.40816213116457223, 0.04545454545454547]}
        """

        check_instance(
            source=NetworkMixin.multigraph_page_rank.__name__,
            instance=graph,
            accepted_types=nx.MultiGraph,
        )
        edge_labels = list(set(data["label"] for _, _, data in graph.edges(data=True)))
        check_iterable_length(
            source=f"{NetworkMixin.multigraph_page_rank.__name__} edge_labels",
            val=len(edge_labels),
            min=1,
        )
        check_str(
            name=f"{NetworkMixin.graph_page_rank.__name__} weights", value=weights
        )
        check_float(name=f"{NetworkMixin.graph_page_rank.__name__} alpha", value=alpha)
        check_int(
            name=f"{NetworkMixin.graph_page_rank.__name__} max_iter",
            value=max_iter,
            min_value=1,
        )
        results = {x: [] for x in list(graph.nodes())}
        for edge_label in edge_labels:
            filtered_graph = nx.Graph(
                graph.edge_subgraph(
                    [
                        (u, v, k)
                        for u, v, k, data in graph.edges(keys=True, data=True)
                        if data.get("label") == edge_label
                    ]
                )
            )
            page_ranks = NetworkMixin().graph_page_rank(
                graph=filtered_graph, weights=weights, alpha=alpha, max_iter=max_iter
            )
            for k, v in page_ranks.items():
                results[k].append(v)

    @staticmethod
    def visualize(graph: Union[nx.Graph, nx.MultiGraph],
                  save_path: Optional[Union[str, os.PathLike]] = None,
                  node_size: Optional[Union[float, Dict[str, float]]] = 25.0,
                  palette: Optional[Union[str, Dict[str, str]]] = "magma",
                  node_shape: Optional[Literal['dot', 'ellipse', 'circle']] = 'dot',
                  smooth_type: Optional[Literal['dynamic', 'continuous', 'discrete', 'diagonalCross', 'straightCross', 'horizontal', 'vertical', 'curvedCW', 'curvedCCW', 'cubicBezier']] = 'dynamic',
                  img_size: Optional[Tuple[int, int]] = (500, 500),
                  seed: Optional[int] = None) -> Union[None, Network]:

        """
        Visualizes a network graph using the vis.js library and saves the result as an HTML file.

        .. raw:: html
           :file: ../../docs/_static/img/network_ex.html


        .. raw:: html

           <iframe src="../../docs/_static/img/network_ex.html" width="100%" height="600px" frameborder="0"></iframe>

        .. note::
           Multi-networks created by ``simba.mixins.network_mixin.create_multigraph`` can be a little messy to look at. Instead,
           creates seperate objects and files with single edges from each time-point.

        :param Union[nx.Graph, nx.MultiGraph] graph: The input graph to be visualized.
        :param Optional[Union[str, os.PathLike]] save_path: The path to save the HTML file. If multi-graph,  pass a directory path. If None, the graph(s) are returned but not saved.
        :param Optional[Union[float, Dict[str, float]]] node_size: The size of nodes. Can be a single float or a dictionary mapping node names to their respective sizes. Default: 25.0.
        :param Optional[Union[str, Dict[str, str]]] palette: The color palette for nodes. Can be a single string representing a palette name or a dictionary mapping node names to their respective colors. Default; magma.
        :param Optional[Tuple[int, int]] img_size: The size of the resulting image in pixels, represented as (width, height). Default: 500x500.
        :param Optional[Literal['dot', 'ellipse', 'circle']] node_shape: The shape of the nodes. Default: `dot`.
        :param Optional[Literal] smooth_type: The dynamics of the interactive graph.

        :example:
        >>> graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
        >>> graph_pg = NetworkMixin.visualize(graph=graph, node_size={'Animal_1': 10, 'Animal_2': 26, 'Animal_3': 50}, save_path='/Users/simon/Downloads/graph.html', node_shape='box', palette='spring')
        >>> multigraph = NetworkMixin().create_multigraph(data={('Animal_1', 'Animal_2'): [0, 0, 0, 6], ('Animal_1', 'Animal_3'): [0, 0, 0, 0], ('Animal_1', 'Animal_4'): [0, 0, 0, 0], ('Animal_1', 'Animal_5'): [0, 0, 0, 0], ('Animal_2', 'Animal_3'): [0, 0, 0, 0], ('Animal_2', 'Animal_4'): [5, 0, 0, 2], ('Animal_2', 'Animal_5'): [0, 0, 0, 0], ('Animal_3', 'Animal_4'): [0, 0, 0, 0], ('Animal_3', 'Animal_5'): [0, 2, 22, 0], ('Animal_4', 'Animal_5'): [0, 0, 0, 0]})
        >>> graph_pg = NetworkMixin.visualize(graph=multigraph, node_size={'Animal_1': 10, 'Animal_2': 26, 'Animal_3': 50, 'Animal_4': 50, 'Animal_5': 50}, save_path='/Users/simon/Downloads/graphs', node_shape='box', palette='spring', smooth_type='diagonalCross')

        """

        check_instance(source=NetworkMixin.visualize.__name__, instance=graph, accepted_types=(nx.MultiGraph, nx.Graph))
        check_instance(source=NetworkMixin.visualize.__name__, instance=node_size, accepted_types=(int, float, dict))
        multi_graph = False
        if graph.is_multigraph(): multi_graph = True
        if multi_graph:
            check_if_dir_exists(in_dir=save_path, source=NetworkMixin.visualize.__name__)
        check_valid_tuple(x=img_size, accepted_lengths=(2,), valid_dtypes=(int,))
        if isinstance(node_size, dict):
            if sorted(graph.nodes) != sorted(list(node_size.keys())):
                raise InvalidInputError(
                    msg=f"node_size keys do not match graph node names: {sorted(graph.nodes)} != {sorted(list(node_size.keys()))}")
            for v in node_size.values():
                check_float(name=f"{NetworkMixin.visualize.__name__} node_size", value=v, min_value=0)
        else:
            check_int(name=f"{NetworkMixin.visualize.__name__} node size", value=node_size, min_value=1)
        if isinstance(palette, dict):
            if sorted(graph.nodes) != sorted(list(palette.keys())):
                raise InvalidInputError(
                    msg=f"palette keys do not match graph node names: {sorted(graph.nodes)} != {sorted(list(node_size.keys()))}")
            for v in palette.values():
                check_valid_hex_color(color_hex=str(v))
            clrs = palette
        else:
            clrs = create_color_palette(pallete_name=palette, as_hex=True, increments=len(list(graph.nodes())))

        if not multi_graph:
            network_graph = Network(f"{img_size[0]}px", f"{img_size[1]}px")
            network_graph.set_edge_smooth(smooth_type)
            if seed is not None:
                pos = nx.circular_layout(graph, 0.4)
            else:
                network_graph.force_atlas_2based()
            for node_cnt, node_name in enumerate(graph):
                if isinstance(node_size, dict):
                    node_node_size = node_size[node_name]
                else:
                    node_node_size = node_size
                if isinstance(palette, dict):
                    node_clr = palette[node_name]
                else:
                    node_clr = clrs[node_cnt]
                if seed is not None:
                    network_graph.add_node(n_id=node_name, shape=node_shape, color=node_clr, size=node_node_size, label=f'Animal {node_name+1}', x=pos[node_name][0] * img_size[0], y=pos[node_name][1] * img_size[1], font={'color': 'white'})
                else:
                    network_graph.add_node(n_id=node_name,
                                           shape=node_shape,
                                           color=node_clr,
                                           size=node_node_size,
                                           label=f'Animal {node_name+1}', font={'color': 'white'})

            network_graph.toggle_physics(False)
            for source, target, edge_attrs in graph.edges(data=True):
                network_graph.add_edge(source, target, value=float(edge_attrs["weight"] +5))
            if save_path is not None:
                network_graph.save_graph(save_path)
            return network_graph

        else:
            results = {}
            edge_labels = list(set(data["label"] for _, _, data in graph.edges(data=True)))
            check_valid_lst(source=f"{NetworkMixin.multigraph_page_rank.__name__} edge_labels", data=edge_labels, min_len=1)
            for edge_label in edge_labels:
                network_graph = Network(f"{img_size[0]}px", f"{img_size[1]}px")
                network_graph.set_edge_smooth(smooth_type)
                network_graph.force_atlas_2based()
                graph_save_path = os.path.join(save_path, f"{edge_label}.html")
                filtered_graph = nx.Graph(graph.edge_subgraph([(u, v, k) for u, v, k, data in graph.edges(keys=True, data=True) if data.get("label") == edge_label]))
                for node_cnt, node_name in enumerate(filtered_graph):
                    if isinstance(node_size, dict):
                        node_node_size = node_size[node_name]
                    else:
                        node_node_size = node_size
                    if isinstance(palette, dict):
                        node_clr = palette[node_name]
                    else:
                        node_clr = clrs[node_cnt]
                    network_graph.add_node(n_id=node_name, shape="dot", color=node_clr, size=node_node_size)
                for source, target, edge_attrs in filtered_graph.edges(data=True):
                    network_graph.add_edge(source, target, value=edge_attrs["weight"])
                if save_path is not None:
                    network_graph.save_graph(graph_save_path)
                results[edge_label] = network_graph
            return results

    @staticmethod
    @jit(nopython=True)
    def simpson_index(x: np.ndarray) -> float:
        """
        Calculate Simpson's diversity index for a given array of values.

        Simpson's diversity index is a measure of diversity that takes into account the number of different categories
        present in the input data as well as the relative abundance of each category. Answer how homogenous a cluster or community is
        for categorical input variable.

        .. math::

           D = \frac{\sum(n(n-1))}{N(N-1)}

        where:
        - \( n \) is the number of individuals of a particular category,
        - \( N \) is the total number of individuals,
        - \( \sum \) represents the sum over all categories.

        :param np.ndarray x: 1-dimensional numpy array containing the values representing categories for which Simpson's index is calculated.
        :return float: Simpson's diversity index value for the input array `x`

        """

        unique_v = np.unique(x)
        n_unique = np.unique(x).shape[0]
        results = np.full((n_unique, 3), np.nan)
        for i in range(unique_v.shape[0]):
            v = unique_v[i]
            cnt = np.argwhere(x == v).flatten().shape[0]
            squared = cnt * (cnt - 1)
            results[i, :] = np.array([v, cnt, squared])
        return (np.sum(results[:, 2])) / (x.shape[0] * (x.shape[0] - 1))

    @staticmethod
    def berger_parker(x: np.ndarray) -> float:
        """
        Berger-Parker index for the given one-dimensional array.
        The Berger-Parker index is a measure of category dominance, calculated as the ratio of
        the frequency of the most abundant category to the total number of observations. Answer how dominated a cluster or community is
        by categorical variable.

        The Berger-Parker index (BP) is calculated using the formula:

        .. math::

           BP = \\frac{f_{\max}}{N}

        where:
        - \( f_{\max} \) is the frequency of the most abundant category,
        - \( N \) is the total number of observations.

        :param np.ndarray x: One-dimensional numpy array containing the values for which the Berger-Parker index is calculated.
        :return float: Berger-Parker index value for the input array `x`

        :example:
        >>> x = np.random.randint(0, 25, (100,)).astype(np.float32)
        >>> z = NetworkMixin.berger_parker(x=x)
        """
        check_valid_array(
            source=f"{NetworkMixin.berger_parker.__name__} x",
            accepted_ndims=(1,),
            data=x,
            accepted_dtypes=(np.float32, np.float64, np.int32, np.int64, np.int8),
        )
        return get_mode(x=x) / x.shape[0]

    @staticmethod
    @jit(nopython=True)
    def shannon_diversity_index(x: np.ndarray) -> float:
        """
        Calculate the Shannon Diversity Index for a given array of categories. The Shannon Diversity Index is a measure of diversity in a
        categorical feature, taking into account both the number of different categories (richness)
        and their relative abundances (evenness). Answer how homogenous a cluster or community is
        for categorical variable. A low value indicates that one or a few categories dominate.

        .. math::

           H = -\sum_{i=1}^{n} (p_i \cdot \log(p_i))

        where:
        - \( p_i \) is the proportion of individuals belonging to the i-th category,
        - \( n \) is the total number of categories.

        :param np.ndarray x: One-dimensional numpy array containing the categories for which the Shannon Diversity Index is calculated.
        :return float: Shannon Diversity Index value for the input array `x`

        :example:
        >>> x = np.random.randint(0, 100, (100, ))
        >>> NetworkMixin.shannon_diversity_index(x=x)
        """

        unique_v = np.unique(x)
        n_unique = np.unique(x).shape[0]
        results = np.full((n_unique,), np.nan)
        for i in range(unique_v.shape[0]):
            v = unique_v[i]
            cnt = np.argwhere(x == v).flatten().shape[0]
            pi = cnt / x.shape[0]
            results[i] = pi * np.log(pi)
        return np.sum(np.abs(results))

    @staticmethod
    def margalef_diversification_index(x: np.array) -> float:
        """
        Calculate the Margalef Diversification Index for a given array of values.

        The Margalef Diversification Index is a measure of category diversity. It quantifies the richness of a community/cluster
        relative to the number of individuals. A high Margalef Diversification Index indicates a high diversity of categories relative to the number of observations.
        A low Margalef Diversification Index suggests a lower diversity of categories relative to the number of observations.

        The Margalef Diversification Index (D) is calculated using the formula:

        .. math::

           D = \\frac{(S - 1)}{\\log(N)}

        where:
        - \( S \) is the number of unique categories,
        - \( N \) is the total number of individuals.

        :param np.array x: One-dimensional numpy array containing nominal values for which the Margalef Diversification Index is calculated.
        :return float: Margalef Diversification Index value for the input array `x`


        :example:
        >>> x = np.random.randint(0, 100, (100,))
        >>> NetworkMixin.margalef_diversification_index(x=x)
        """
        check_valid_array(
            source=f"{NetworkMixin.margalef_diversification_index.__name__} x",
            accepted_ndims=(1,),
            data=x,
            accepted_dtypes=(np.float32, np.float64, np.int32, np.int64, np.int8),
            min_axis_0=2,
        )
        n_unique = np.unique(x).shape[0]
        return (n_unique - 1) / np.log(x.shape[0])

    @staticmethod
    def menhinicks_index(x: np.array) -> float:
        """
        Calculate the Menhinick's Index for a given array of values.

        Menhinick's Index is a measure of category richness. It quantifies the number of categories relative to the square root of the total number of observations.
        A high Menhinick's Index suggests a high diversity of categories relative to the number of observations.
        A low Menhinick's Index indicates a lower diversity of categories relative to the number of observations.

        Menhinick's Index (D) is calculated using the formula:

        .. math::

           D = \\frac{S}{\\sqrt{N}}

        where:
        - \( S \) is the number of unique categories,
        - \( N \) is the total number of observations.

         :param np.array x: One-dimensional numpy array containing the integer values representing nominal values for which Menhinick's Index is calculated.
         :return float: Menhinick's Index value for the input array `x`

        :example:
        >>> x = np.random.randint(0, 5, (1000,))
        >>> NetworkMixin.menhinicks_index(x=x)
        """
        check_valid_array(
            source=f"{NetworkMixin.menhinicks_index.__name__} x",
            accepted_ndims=(1,),
            data=x,
            accepted_dtypes=(np.float32, np.float64, np.int32, np.int64, np.int8),
            min_axis_0=2,
        )
        return np.unique(x).shape[0] / np.sqrt(x.shape[0])

    @staticmethod
    def brillouins_index(x: np.array) -> float:
        """
        Calculate Brillouin's Diversity Index for a given array of values.

        Brillouin's Diversity Index is a measure of cluster/community diversity that accounts for both richness
        and evenness of distribution.

        Brillouin's Diversity Index (H) is calculated using the formula:

        .. math::

           H = \\frac{1}{\\log(S)} \\sum_{i=1}^{S} \\frac{N_i(N_i - 1)}{n(n-1)}

        where:
        - \( H \) is Brillouin's Diversity Index,
        - \( S \) is the total number of unique species,
        - \( N_i \) is the count of individuals in the i-th species,
        - \( n \) is the total number of individuals.

        :param np.array x: One-dimensional numpy array containing the values for which Brillouin's Index is calculated.
        :return float: Brillouin's Diversity Index value for the input array `x`

        :example:
        >>> x = np.random.randint(0, 10, (100,))
        >>> NetworkMixin.brillouins_index(x)
        """

        check_valid_array(
            source=f"{NetworkMixin.brillouins_index.__name__} x",
            accepted_ndims=(1,),
            data=x,
            accepted_dtypes=(np.float32, np.float64, np.int32, np.int64, np.int8),
            min_axis_0=2,
        )
        n_total = x.shape[0]
        n_unique = np.unique(x, return_counts=True)[1]
        if n_unique.shape[0] == 1:
            return 1.0
        else:
            S = len(n_unique)
            h = 0
            for count in n_unique:
                h += count * (count - 1)
            h /= n_total * (n_total - 1)
            h /= np.log(S)
            return h

    @staticmethod
    def sorensen_dice_coefficient(x: np.ndarray, y: np.ndarray) -> float:
        """
        Calculate Sørensen's Similarity Index between two communities/clusters.

        The Sørensen similarity index, also known as the overlap index, quantifies the overlap between two populations by comparing the number of shared categories to the total number of categories in both populations. It ranges from zero, indicating no overlap, to one, representing perfect overlap

        Sørensen's Similarity Index (S) is calculated using the formula:

        .. math::

            S = \\frac{2 \times |X \cap Y|}{|X| + |Y|}

        where:
        - \( S \) is Sørensen's Similarity Index,
        - \( X \) and \( Y \) are the sets representing the categories in the first and second communities, respectively,
        - \( |X \cap Y| \) is the number of shared categories between the two communities,
        - \( |X| \) and \( |Y| \) are the total number of categories in the first and second communities, respectively.


        :param x: 1D numpy array with nominal values for the first cluster/community.
        :param y: 1D numpy array with nominal values for the second cluster/community.
        :return: Sørensen's Similarity Index between x and y.

        :example:
        >>> x = np.random.randint(0, 10, (100,))
        >>> y = np.random.randint(0, 10, (100,))
        >>> NetworkMixin.sorensen_dice_coefficient(x=x, y=y)
        """

        check_valid_array(
            source=f"{NetworkMixin.sorensen_dice_coefficient.__name__} x",
            accepted_ndims=(1,),
            data=x,
            accepted_dtypes=(np.int32, np.int64, np.int8, int),
            min_axis_0=2,
        )
        check_valid_array(
            source=f"{NetworkMixin.sorensen_dice_coefficient.__name__} y",
            accepted_ndims=(1,),
            data=y,
            accepted_dtypes=(np.int32, np.int64, np.int8, int),
            min_axis_0=2,
        )
        x, y = set(x), set(y)
        return 2 * len(x.intersection(y)) / (len(x) + len(y))


# graph = NetworkMixin.create_graph(
#     {
#         ("Animal_1", "Animal_2"): 0.0,
#         ("Animal_1", "Animal_3"): 0.0,
#         ("Animal_1", "Animal_4"): 0.0,
#         ("Animal_1", "Animal_5"): 0.1,
#         ("Animal_2", "Animal_3"): 1.0,
#         ("Animal_2", "Animal_4"): 1.0,
#         ("Animal_2", "Animal_5"): 1.5,
#         ("Animal_3", "Animal_4"): 1.0,
#         ("Animal_3", "Animal_5"): 1.0,
#         ("Animal_4", "Animal_5"): 1.0,
#     }
# )
#
# G = nx.Graph(graph)
#
# NetworkMixin().graph_current_flow_closeness_centrality(graph=graph)
# #
#
# communities = list(nx.algorithms.community.(G, weight='weight'))
#
# partition = nx.algorithms.community.kernighan_lin_bisection(G, weight='weight')
# partition = nx.algorithms.community.greedy_modularity_communities(G, weight='weight')
# partition = list(nx.algorithms.community.girvan_newman(G))
#
# tuple(sorted(c) for c in next(partition))

# graph_data = {
#     ('Animal_1', 'Animal_2'): 0.0,
#     ('Animal_1', 'Animal_3'): 0.0,
#     ('Animal_1', 'Animal_4'): 0.0,
#     ('Animal_1', 'Animal_5'): 0.0,
#     ('Animal_2', 'Animal_3'): 1.5,  # Increase weight
#     ('Animal_2', 'Animal_4'): 1.0,  # Increase weight
#     ('Animal_2', 'Animal_5'): 1.5,  # Increase weight
#     ('Animal_3', 'Animal_4'): 2.9,  # Increase weight
#     ('Animal_3', 'Animal_5'): 2.0,  # Increase weight
#     ('Animal_4', 'Animal_5'): 1.0   # Decrease weight
# }

# result = list(NetworkMixin().asyn_lpa_communities(graph=graph, weight='weight'))
#
#
# for i in list(result):
#     print(i)
#


# graph_pg = NetworkMixin().graph_page_rank(graph=graph)
# graph_clrs = find_ranked_colors(data=graph_pg, palette='jet', as_hex=True)
# NetworkMixin().visualize(graph=graph,
#                          save_path='/Users/simon/Desktop/envs/troubleshooting/ARES_data/Termite Test/project/project_data/network_html/graph_101.html',
#                          palette=graph_clrs)


# multigraph = NetworkMixin().create_multigraph(data={('Animal_1', 'Animal_2'): [0, 0, 0, 6], ('Animal_1', 'Animal_3'): [0, 0, 0, 0], ('Animal_1', 'Animal_4'): [0, 0, 0, 0], ('Animal_1', 'Animal_5'): [0, 0, 0, 0], ('Animal_2', 'Animal_3'): [0, 0, 0, 0], ('Animal_2', 'Animal_4'): [5, 0, 0, 2], ('Animal_2', 'Animal_5'): [0, 0, 0, 0], ('Animal_3', 'Animal_4'): [0, 0, 0, 0], ('Animal_3', 'Animal_5'): [0, 2, 22, 0], ('Animal_4', 'Animal_5'): [0, 0, 0, 0]})
#
#
# #graph_pg = NetworkMixin().graph_page_rank(graph=graph)
# graph_clrs = find_ranked_colors(data=graph_pg, palette='jet', as_hex=True)
#
# NetworkMixin().visualize(graph=graph,
#                          save_path='/Users/simon/Desktop/envs/troubleshooting/ARES_data/Termite Test/project/project_data/network_html.html',
#                          palette=graph_clrs)


# multigraph = NetworkMixin().create_multigraph(data={('Animal_1', 'Animal_2'): [0, 0, 0, 6], ('Animal_1', 'Animal_3'): [0, 0, 0, 0], ('Animal_1', 'Animal_4'): [0, 0, 0, 0], ('Animal_1', 'Animal_5'): [0, 0, 0, 0], ('Animal_2', 'Animal_3'): [0, 0, 0, 0], ('Animal_2', 'Animal_4'): [5, 0, 0, 2], ('Animal_2', 'Animal_5'): [0, 0, 0, 0], ('Animal_3', 'Animal_4'): [0, 0, 0, 0], ('Animal_3', 'Animal_5'): [0, 2, 22, 0], ('Animal_4', 'Animal_5'): [0, 0, 0, 0]})
# NetworkMixin().visualize(graph=multigraph, save_path='/Users/simon/Desktop/envs/troubleshooting/ARES_data/Termite Test/project/project_data/network_html')


# graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
# NetworkMixin().graph_page_rank(graph=graph)
# multigraph = NetworkMixin().create_multigraph(data=data)


# multigraph = NetworkMixin().create_multigraph(data=data)
#
# #multigraph.edges.data()
#
# NetworkMixin().visualize(graph=multigraph, save_dir='/Users/simon/Desktop/envs/troubleshooting/ARES_data/Termite Test/project/project_data/network_html')


#
# for video_name, G in self.graphs.items():
#     if node_colors is not None:
#         clrs = create_single_color_lst(pallete_name='magma', increments=len(self.animal_names), as_hex=True)
#         node_colors = {k: v for k, v in sorted(node_colors.items(), key=lambda item: item[1], reverse=True)}
#         for node_cnt, node_name in enumerate(node_colors.keys()):
#             node_colors[node_name] = clrs[node_cnt]
#
#     network_graph = Network(style_attr['size'][0], style_attr['size'][1])
#     for node_name, node_attrs in G.nodes(data=True):
#         network_graph.add_node(node_name, color=node_colors[node_name])
#
#     for source, target, edge_attrs in G.edges(data=True):
#         edge_attrs['value'] = edge_attrs['weight']
#         network_graph.add_edge(str(source), str(target), **edge_attrs)
#
#     network_graph.save_graph('nx.html')
#


#
#
# data = {('Animal_1', 'Animal_2'): 0.0,
#         ('Animal_1', 'Animal_3'): 0.2,
#         ('Animal_1', 'Animal_4'): 0.3,
#         ('Animal_2', 'Animal_3'): 0.9,
#         ('Animal_2', 'Animal_4'): 1.0,
#         ('Animal_3', 'Animal_4'): 0.4}
# graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
# NetworkMixin().graph_current_flow_closeness_centrality(graph=graph)
#


#
#
# data = {('Animal_1', 'Animal_2'): [0, 0, 0, 6],
#         ('Animal_1', 'Animal_3'): [0, 0, 0, 0],
#         ('Animal_1', 'Animal_4'): [0, 0, 0, 0],
#         ('Animal_1', 'Animal_5'): [0, 0, 0, 0],
#         ('Animal_2', 'Animal_3'): [0, 0, 0, 0],
#         ('Animal_2', 'Animal_4'): [5, 0, 0, 2],
#         ('Animal_2', 'Animal_5'): [0, 0, 0, 0],
#         ('Animal_3', 'Animal_4'): [0, 0, 0, 0],
#         ('Animal_3', 'Animal_5'): [0, 2, 22, 0],
#         ('Animal_4', 'Animal_5'): [0, 0, 0, 0]}


# graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
# #NetworkMixin().graph_page_rank(graph=graph)
# multigraph = NetworkMixin().create_multigraph(data=data)
#
# #multigraph.edges.data()
#
# NetworkMixin().visualize(graph=multigraph, save_dir='/Users/simon/Desktop/envs/troubleshooting/ARES_data/Termite Test/project/project_data/network_html')


# NetworkMixin().multigraph_page_rank(graph=multigraph)


# NetworkMixin.page_rank(graph=graph)
# NetworkMixin.katz_centrality(graph=graph)
# NetworkMixin.current_flow_closeness_centrality(graph=graph)



__author__ = "Simon Nilsson"

import os
from copy import deepcopy
from typing import Optional, Union

import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log, check_float,
    check_if_dir_exists, check_if_keys_exist_in_dict)
from simba.utils.data import plug_holes_shortest_bout
from simba.utils.enums import TagNames
from simba.utils.errors import NoFilesFoundError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_df, write_df)
from simba.utils.warnings import NoFileFoundWarning


class InferenceBatch(TrainModelMixin, ConfigReader):
    """
    Run classifier inference on all files with the ``project_folder/csv/features_extracted`` directory.
    Results are stored in the ``project_folder/csv/machine_results`` directory of the SimBA project.

    .. note::
       To compute aggregate statistics from the output of this class, see :func:`simba.data_processors.agg_clf_calculator.AggregateClfCalculator`

    :param Union[str, os.PathLike] config_path: path to SimBA project config file in Configparser format.
    :param Optional[Union[str, os.PathLike]] features_dir: Optional directory containing featurized files in CSV or parquet format. If None, then the `project_folder/csv/features_extracted` directory of the project will be used.
    :param Optional[Union[str, os.PathLike]] save_dir: Optional directory to save the data for the analyzed videos. If None, then the `project_folder/csv/machine_results` directory of the project will be used.

    :example I:
    >>> inferencer = InferenceBatch(config_path='MyConfigPath')
    >>> inferencer.run()

    :example II:
    >>> inferencer = InferenceBatch(config_path=r"D:\troubleshooting\mitra\project_folder\project_config.ini", features_dir=r"D:\troubleshooting\mitra\project_folder\videos\bg_removed\rotated\tail_features\APPENDED")
    >>> inferencer.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 features_dir: Optional[Union[str, os.PathLike]] = None,
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 minimum_bout_length: Optional[int] = None):

        ConfigReader.__init__(self, config_path=config_path)
        if features_dir is not None:
            check_if_dir_exists(in_dir=features_dir, source=self.__class__.__name__)
            self.features_dir = features_dir
            self.feature_file_paths = find_files_of_filetypes_in_directory(directory=self.features_dir, extensions=[f'.{self.file_type}'], raise_warning=False, raise_error=False)
        if save_dir is not None:
            check_if_dir_exists(in_dir=features_dir, source=self.__class__.__name__)
            self.save_dir = save_dir
        else:
            self.save_dir = self.machine_results_dir
        TrainModelMixin.__init__(self)
        log_event(logger_name=str(self.__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        if len(self.feature_file_paths) == 0:
            raise NoFilesFoundError(f"Zero files found in the {self.features_dir}. Create features before running classifier.", source=self.__class__.__name__,)
        print(f"Analyzing {len(self.feature_file_paths)} file(s) with {self.clf_cnt} classifier(s)")
        self.timer = SimbaTimer(start=True)
        self.model_dict = self.get_model_info(config=self.config, model_cnt=self.clf_cnt)
        # clf_paths = [v["model_path"] for k, v in self.model_dict.items()]
        # invalid_clf_paths = [k for k, v in self.model_dict.items() if not clf_paths]
        # if len(invalid_clf_paths) == len(clf_paths):
        #     raise NoFilesFoundError(msg=f"None of the classifier paths specified in the config file exists: {clf_paths}", source=self.__class__.__name__)
        # if len(invalid_clf_paths) > 0:
        #     for clf_name in invalid_clf_paths:
        #         NoFileFoundWarning(msg=f'SKIPPING CLASSIFIER {self.model_dict["model_name"]}. The classifier model file {self.model_dict[clf_name]} could not be found.', source=self.__class__.__name__)
        #         self.model_dict = {k: v for k, v in self.model_dict.items() if k != clf_name}

    def run(self):
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.feature_file_paths)
        for file_cnt, file_path in enumerate(self.feature_file_paths):
            video_timer = SimbaTimer(start=True)
            _, file_name, _ = get_fn_ext(file_path)
            print(f"Analyzing video {file_name}... (Video {file_cnt+1}/{len(self.feature_file_paths)})")
            file_save_path = os.path.join(self.save_dir, f"{file_name}.{self.file_type}")
            in_df = read_df(file_path, self.file_type)
            x_df = self.drop_bp_cords(df=in_df).astype(np.float32)
            self.check_df_dataset_integrity(df=x_df, logs_path=self.logs_path, file_name=file_name)
            _, px_per_mm, fps = self.read_video_info(video_name=file_name, raise_error=False)
            out_df = deepcopy(in_df)
            for m, m_hyp in self.model_dict.items():
                check_if_keys_exist_in_dict(data=m_hyp, key=['model_path', "model_name", "threshold", 'minimum_bout_length'], name=f'classifier {m}', raise_error=False)
                if not os.path.isfile(m_hyp["model_path"]):
                    NoFileFoundWarning(msg=f'SKIPPING CLASSIFIER {m} for video {file_name}. The classifier model file {m_hyp["model_path"]} could not be found.', source=self.__class__.__name__)
                    continue
                    #raise NoFilesFoundError(msg=f"{m_hyp['model_path']} is not a VALID model file path", source=self.__class__.__name__)
                probability_column = f"Probability_{m_hyp['model_name']}"
                clf = self.read_pickle(file_path=m_hyp["model_path"])
                out_df[probability_column] = self.clf_predict_proba(clf=clf, x_df=x_df, data_path=file_path, model_name=m_hyp["model_name"])
                out_df[m_hyp["model_name"]] = np.where(out_df[probability_column] > m_hyp["threshold"], 1, 0)
                if int(m_hyp["minimum_bout_length"]) > 0:
                    print(f'Correcting minimum bouts in video {file_name} and classifier {m_hyp["model_name"]} ({m_hyp["minimum_bout_length"]}ms)...')
                    out_df = plug_holes_shortest_bout(data_df=out_df, clf_name=m_hyp["model_name"], fps=fps, shortest_bout=m_hyp["minimum_bout_length"])
            write_df(out_df, self.file_type, file_save_path)
            video_timer.stop_timer()
            print(f"Predictions created for {file_name} (elapsed time: {video_timer.elapsed_time_str}) ...")
        self.timer.stop_timer()
        stdout_success(msg=f"Machine predictions complete. Files saved in {self.save_dir} directory", elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)

# test = InferenceBatch(config_path=r"C:\troubleshooting\Top_down_old\project_folder\project_config.ini")
# test.run()


# test = InferenceBatch(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                       features_dir=r'C:\troubleshooting\mitra\project_folder\videos\additional\bg_removed\rotated\tail_features_additional\APPENDED')
# test.run()



# test = InferenceBatch(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
# test.run()





# test = InferenceBatch(config_path=r"D:\troubleshooting\mitra\project_folder\project_config.ini",
#                       features_dir=r'D:\troubleshooting\mitra\project_folder\videos\bg_removed\rotated\laying_down_features\APPENDED')
# test.run()

# test = InferenceBatch(config_path=r"D:\troubleshooting\mitra\project_folder\project_config.ini", features_dir=r'D:\troubleshooting\mitra\project_folder\videos\bg_removed\rotated\tail_features\APPENDED')
# test.run()


# test = InferenceBatch(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/levi/project_folder/project_config.ini')
# test.run()

# test = InferenceBatch(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.run()

# test = RunModel(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.run_models()



import os
from typing import Optional, Union

import numpy as np
import pandas as pd
from numba import typed

from simba.mixins.circular_statistics import CircularStatisticsMixin
from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.mixins.timeseries_features_mixin import TimeseriesFeatureMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log, check_if_dir_exists,
    check_int, check_str, check_valid_dataframe)
from simba.utils.data import detect_bouts, plug_holes_shortest_bout
from simba.utils.enums import Formats
from simba.utils.printing import stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_df, read_video_info)

CIRCLING = 'CIRCLING'

class CirclingDetector(ConfigReader):

    """
    Detect circling using heuristic rules.

    .. important::

        Circling is detected as `present` when **the circular range of the animal is above the specied circular range threshold within the specified preceding time threshold** AND
        **the movement of the animal (defined as the sum of the center movement) is above the specified movement threshold within the specified preceding time threshold.**

        Circling is detected as `absent` when not present.

    .. note::

       We pass the names of the left ear, right ear, and nose, as the method will use these body-parts to compute the direction of the animal in degrees.

    :param Union[str, os.PathLike] data_dir: Path to directory containing pose-estimated body-part data in CSV format.
    :param Union[str, os.PathLike] config_path: Path to SimBA project config file.
    :param Optional[str] nose_name: The name of the pose-estimated nose body-part. Defaults to 'nose'.
    :param Optional[str] left_ear_name: The name of the pose-estimated left ear body-part. Defaults to 'left_ear'.
    :param Optional[str] right_ear_name: The name of the pose-estimated right ear body-part. Defaults to 'right_ear'.
    :param Optional[str] tail_base_name: The name of the pose-estimated tail base body-part. Defaults to 'tail_base'.
    :param Optional[str] center_name: The name of the pose-estimated center body-part. Defaults to 'center'.
    :param Optional[int] time_threshold: The time window in preceding seconds in which to evaluate the animals circular range. Default: 10.
    :param Optional[int] circular_range_threshold: A value in degrees, between 0-360.
    :param Optional[int] movement_threshold: A movement threshold in millimeters. Defaults to 60.
    :param Optional[Union[str, os.PathLike]] save_dir: Directory where to store the results. If None, then results are stored in the ``logs`` directory of the SimBA project.

    References
    ----------
    .. [1] Sabnis et al., Visual detection of seizures in mice using supervised machine learning, `biorxiv`, doi: https://doi.org/10.1101/2024.05.29.596520.

    :example:
    >>> CirclingDetector(data_dir=r'D:\troubleshooting\mitra\project_folder\csv\outlier_corrected_movement_location', config_path=r"D:\troubleshooting\mitra\project_folder\project_config.ini")
    """

    def __init__(self,
                 data_dir: Union[str, os.PathLike],
                 config_path: Union[str, os.PathLike],
                 nose_name: Optional[str] = 'nose',
                 left_ear_name: Optional[str] = 'left_ear',
                 right_ear_name: Optional[str] = 'right_ear',
                 tail_base_name: Optional[str] = 'tail_base',
                 center_name: Optional[str] = 'center',
                 time_threshold: Optional[int] = 10,
                 circular_range_threshold: Optional[int] = 320,
                 movement_threshold: Optional[int] = 60,
                 save_dir: Optional[Union[str, os.PathLike]] = None):

        check_if_dir_exists(in_dir=data_dir)
        for bp_name in [nose_name, left_ear_name, right_ear_name, tail_base_name]: check_str(name='body part name', value=bp_name, allow_blank=False)
        self.data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=['.csv'])
        ConfigReader.__init__(self, config_path=config_path, read_video_info=True, create_logger=False)
        self.nose_heads = [f'{nose_name}_x'.lower(), f'{nose_name}_y'.lower()]
        self.left_ear_heads = [f'{left_ear_name}_x'.lower(), f'{left_ear_name}_y'.lower()]
        self.right_ear_heads = [f'{right_ear_name}_x'.lower(), f'{right_ear_name}_y'.lower()]
        self.center_heads = [f'{center_name}_x'.lower(), f'{center_name}_y'.lower()]
        self.required_field = self.nose_heads + self.left_ear_heads + self.right_ear_heads
        self.save_dir = save_dir
        if self.save_dir is None:
            self.save_dir = os.path.join(self.logs_path, f'circling_data_{self.datetime}')
            os.makedirs(self.save_dir)
        else:
            check_if_dir_exists(in_dir=self.save_dir)
        self.time_threshold, self.circular_range_threshold, self.movement_threshold = time_threshold, circular_range_threshold, movement_threshold

    def run(self):
        agg_results = pd.DataFrame(columns=['VIDEO', 'CIRCLING FRAMES', 'CIRCLING TIME (S)', 'CIRCLING BOUT COUNTS', 'CIRCLING PCT OF SESSION', 'VIDEO TOTAL FRAMES', 'VIDEO TOTAL TIME (S)'])
        agg_results_path = os.path.join(self.save_dir, 'aggregate_circling_results.csv')
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        for file_cnt, file_path in enumerate(self.data_paths):
            video_name = get_fn_ext(filepath=file_path)[1]
            print(f'Analyzing {video_name} ({file_cnt+1}/{len(self.data_paths)})...')
            save_file_path = os.path.join(self.save_dir, f'{video_name}.csv')
            df = read_df(file_path=file_path, file_type='csv').reset_index(drop=True)
            _, px_per_mm, fps = read_video_info(video_info_df=self.video_info_df, video_name=video_name)
            df.columns = [str(x).lower() for x in df.columns]
            check_valid_dataframe(df=df, valid_dtypes=Formats.NUMERIC_DTYPES.value, required_fields=self.required_field)

            nose_arr = df[self.nose_heads].values.astype(np.float32)
            left_ear_arr = df[self.left_ear_heads].values.astype(np.float32)
            right_ear_arr = df[self.right_ear_heads].values.astype(np.float32)

            center_shifted = FeatureExtractionMixin.create_shifted_df(df[self.center_heads])
            center_1, center_2 = center_shifted.iloc[:, 0:2].values, center_shifted.iloc[:, 2:4].values

            angle_degrees = CircularStatisticsMixin().direction_three_bps(nose_loc=nose_arr, left_ear_loc=left_ear_arr, right_ear_loc=right_ear_arr).astype(np.float32)
            sliding_circular_range = CircularStatisticsMixin().sliding_circular_range(data=angle_degrees, time_windows=np.array([self.time_threshold], dtype=np.float64), fps=int(fps)).flatten()
            movement = FeatureExtractionMixin.euclidean_distance(bp_1_x=center_1[:, 0].flatten(), bp_2_x=center_2[:, 0].flatten(), bp_1_y=center_1[:, 1].flatten(), bp_2_y=center_2[:, 1].flatten(), px_per_mm=2.15)
            movement_sum = TimeseriesFeatureMixin.sliding_descriptive_statistics(data=movement.astype(np.float32), window_sizes=np.array([self.time_threshold], dtype=np.float64), sample_rate=fps, statistics=typed.List(["sum"])).astype(np.int32)[0].flatten()

            circling_idx = np.argwhere(sliding_circular_range >= self.circular_range_threshold).astype(np.int32).flatten()
            movement_idx = np.argwhere(movement_sum >= self.movement_threshold).astype(np.int32).flatten()
            circling_idx = [x for x in movement_idx if x in circling_idx]
            df[CIRCLING] = 0
            df.loc[circling_idx, CIRCLING] = 1
            bouts = detect_bouts(data_df=df, target_lst=[CIRCLING], fps=fps)
            df = plug_holes_shortest_bout(data_df=df, clf_name=CIRCLING, fps=fps, shortest_bout=100)
            df.to_csv(save_file_path)
            agg_results.loc[len(agg_results)] = [video_name, len(circling_idx), round(len(circling_idx) / fps, 4), len(bouts), round((len(circling_idx) / len(df)) * 100, 4), len(df), round(len(df)/fps, 2) ]

        agg_results.to_csv(agg_results_path)
        stdout_success(msg=f'Results saved in {self.save_dir} directory.')

#
#
# detector = CirclingDetector(data_dir=r'C:\troubleshooting\mitra\project_folder\csv\outlier_corrected_movement_location',
#                             config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
# detector.run()

